# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os
import sys
import itertools

import numpy as np

from ._libtoast import Logger

from .pshmem import MPIShared, MPILock

use_mpi = None
MPI = None

if use_mpi is None:
    # See if the user has explicitly disabled MPI.
    if "MPI_DISABLE" in os.environ:
        use_mpi = False
    else:
        # Special handling for running on a NERSC login node.  This is for convenience.
        # The same behavior could be implemented with environment variables set in a
        # shell resource file.
        at_nersc = False
        if "NERSC_HOST" in os.environ:
            at_nersc = True
        in_slurm = False
        if "SLURM_JOB_NAME" in os.environ:
            in_slurm = True
        if (not at_nersc) or in_slurm:
            try:
                import mpi4py.MPI as MPI

                use_mpi = True
            except:
                # There could be many possible exceptions raised...
                from ._libtoast import Logger

                log = Logger.get()
                log.info("mpi4py not found- using serial operations only")
                use_mpi = False


def get_world():
    """Retrieve the default world communicator and its properties.

    If MPI is enabled, this returns MPI.COMM_WORLD and the process rank and number of
    processes.  If MPI is disabled, this returns None for the communicator, zero
    for the rank, and one for the number of processes.

    Returns:
        (tuple):  The communicator, number of processes, and rank.

    """
    rank = 0
    procs = 1
    world = None
    if use_mpi:
        world = MPI.COMM_WORLD
        rank = world.rank
        procs = world.size
    return world, procs, rank


class Comm(object):
    """Class which represents a two-level hierarchy of MPI communicators.

    A Comm object splits the full set of processes into groups of size
    "group".  If group_size does not divide evenly into the size of the given
    communicator, then those processes remain idle.

    A Comm object stores three MPI communicators:  The "world" communicator
    given here, which contains all processes to consider, a "group"
    communicator (one per group), and a "rank" communicator which contains the
    processes with the same group-rank across all groups.

    If MPI is not enabled, then all communicators are set to None.

    Args:
        world (mpi4py.MPI.Comm): the MPI communicator containing all processes.
        group (int): the size of each process group.

    """

    def __init__(self, world=None, groupsize=0):
        log = Logger.get()
        if world is None:
            if use_mpi:
                # Default is COMM_WORLD
                world = MPI.COMM_WORLD
            else:
                # MPI is disabled, leave world as None.
                pass
        else:
            if use_mpi:
                # We were passed a communicator to use. Check that it is
                # actually a communicator, otherwise fall back to COMM_WORLD.
                if not isinstance(world, MPI.Comm):
                    log.warning(
                        "Specified world communicator is not a valid "
                        "mpi4py.MPI.Comm object.  Using COMM_WORLD."
                    )
                    world = MPI.COMM_WORLD
            else:
                log.warning(
                    "World communicator specified even though "
                    "MPI is disabled.  Ignoring this constructor "
                    "argument."
                )
                world = None
            # Special case, MPI available but the user want a serial
            # data object
            if world == MPI.COMM_SELF:
                world = None

        self._wcomm = world
        self._wrank = 0
        self._wsize = 1
        if self._wcomm is not None:
            self._wrank = self._wcomm.rank
            self._wsize = self._wcomm.size

        self._gsize = groupsize

        if (self._gsize < 0) or (self._gsize > self._wsize):
            log.warning(
                "Invalid groupsize ({}).  Should be between {} "
                "and {}.  Using single process group instead.".format(
                    groupsize, 0, self._wsize
                )
            )
            self._gsize = 0

        if self._gsize == 0:
            self._gsize = self._wsize

        self._ngroups = self._wsize // self._gsize

        if self._ngroups * self._gsize != self._wsize:
            msg = (
                "World communicator size ({}) is not evenly divisible "
                "by requested group size ({}).".format(self._wsize, self._gsize)
            )
            log.error(msg)
            raise RuntimeError(msg)

        self._group = self._wrank // self._gsize
        self._grank = self._wrank % self._gsize

        if self._ngroups == 1:
            # We just have one group with all processes.
            self._gcomm = self._wcomm
            if use_mpi:
                self._rcomm = MPI.COMM_SELF
            else:
                self._rcomm = None
        else:
            # We need to split the communicator.  This code is never executed
            # unless MPI is enabled and we have multiple groups.
            self._gcomm = self._wcomm.Split(self._group, self._grank)
            self._rcomm = self._wcomm.Split(self._grank, self._group)

    @property
    def world_size(self):
        """The size of the world communicator."""
        return self._wsize

    @property
    def world_rank(self):
        """The rank of this process in the world communicator."""
        return self._wrank

    @property
    def ngroups(self):
        """The number of process groups."""
        return self._ngroups

    @property
    def group(self):
        """The group containing this process."""
        return self._group

    @property
    def group_size(self):
        """The size of the group containing this process."""
        return self._gsize

    @property
    def group_rank(self):
        """The rank of this process in the group communicator."""
        return self._grank

    @property
    def comm_world(self):
        """The world communicator."""
        return self._wcomm

    @property
    def comm_group(self):
        """The communicator shared by processes within this group."""
        return self._gcomm

    @property
    def comm_rank(self):
        """The communicator shared by processes with the same group_rank."""
        return self._rcomm

    def __repr__(self):
        lines = [
            "  World MPI communicator = {}".format(self._wcomm),
            "  World MPI size = {}".format(self._wsize),
            "  World MPI rank = {}".format(self._wrank),
            "  Group MPI communicator = {}".format(self._gcomm),
            "  Group MPI size = {}".format(self._gsize),
            "  Group MPI rank = {}".format(self._grank),
            "  Rank MPI communicator = {}".format(self._rcomm),
        ]
        return "<toast.Comm\n{}\n>".format("\n".join(lines))



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..map import DistPixels
from ..todmap import OpSimPySM, OpSimScan, OpMadam, OpSimConviqt, OpSimWeightedConviqt


def add_sky_signal_args(parser):
    # The sky signal flag may be already added
    try:
        parser.add_argument(
            "--sky",
            required=False,
            action="store_true",
            help="Add simulated sky signal",
            dest="simulate_sky",
        )
        parser.add_argument(
            "--simulate-sky",
            required=False,
            action="store_true",
            help="Add simulated sky",
            dest="simulate_sky",
        )
        parser.add_argument(
            "--no-sky",
            required=False,
            action="store_false",
            help="Do not add simulated sky",
            dest="simulate_sky",
        )
        parser.add_argument(
            "--no-simulate-sky",
            required=False,
            action="store_false",
            help="Do not add simulated sky",
            dest="simulate_sky",
        )
        parser.set_defaults(simulate_sky=True)
    except argparse.ArgumentError:
        pass
    return


def add_sky_map_args(parser):
    """Add the sky arguments"""

    parser.add_argument(
        "--input-map",
        required=False,
        help="Input map for signal.  You can use Python formatting for Monte "
        "Carlo realization, {mc:04i}, and detector name, {detector}",
    )

    # The nside may already be added
    try:
        parser.add_argument(
            "--nside", required=False, default=512, type=np.int, help="Healpix NSIDE"
        )
    except argparse.ArgumentError:
        pass
    # The coordinate system may already be added
    try:
        parser.add_argument(
            "--coord", required=False, default="C", help="Sky coordinate system [C,E,G]"
        )
    except argparse.ArgumentError:
        pass
    add_sky_signal_args(parser)

    return


def add_pysm_args(parser):
    """Add the sky arguments"""

    parser.add_argument(
        "--pysm-model",
        required=False,
        help="Comma separated models for on-the-fly PySM "
        'simulation, e.g. "s1,d6,f1,a2"',
    )

    parser.add_argument(
        "--pysm-apply-beam",
        required=False,
        action="store_true",
        help="Convolve sky with detector beam",
        dest="pysm_apply_beam",
    )
    parser.add_argument(
        "--no-pysm-apply-beam",
        required=False,
        action="store_false",
        help="Do not convolve sky with detector beam.",
        dest="pysm_apply_beam",
    )
    parser.set_defaults(pysm_apply_beam=True)

    parser.add_argument(
        "--pysm-precomputed-cmb-K_CMB",
        required=False,
        help="Precomputed CMB map for PySM in K_CMB"
        'it overrides any model defined in pysm_model"',
    )

    parser.add_argument(
        "--pysm-mpi-comm",
        required=False,
        help="MPI communicator used by the PySM operator, either 'rank' or 'group'",
        dest="pysm_mpi_comm",
    )
    parser.set_defaults(pysm_mpi_comm="group")

    # The nside may already be added
    try:
        parser.add_argument(
            "--nside", required=False, default=512, type=np.int, help="Healpix NSIDE"
        )
    except argparse.ArgumentError:
        pass
    # The coordinate system may already be added
    try:
        parser.add_argument(
            "--coord", required=False, default="C", help="Sky coordinate system [C,E,G]"
        )
    except argparse.ArgumentError:
        pass

    add_sky_signal_args(parser)

    return


def add_conviqt_args(parser):
    """Add arguments for synthesizing signal with libConviqt"""
    parser.add_argument(
        "--conviqt-sky-file",
        required=False,
        help="Path to sky alm files. Tag {detector} will be "
        "replaced with detector name.",
    )
    parser.add_argument(
        "--conviqt-lmax",
        required=False,
        type=np.int,
        help="Simulation lmax.  May not exceed the expansion order in conviqt-sky-file",
    )
    parser.add_argument(
        "--conviqt-fwhm",
        required=False,
        type=np.float,
        help="Sky fwhm [arcmin] to deconvolve",
    )
    parser.add_argument(
        "--conviqt-beam-file",
        required=False,
        help="Path to beam alm files. Tag {detector} will be "
        "replaced with detector name.",
    )
    parser.add_argument(
        "--conviqt-beam-mmax",
        required=False,
        type=np.int,
        help="Beam mmax.  May not exceed the expansion order in conviqt-beam-file",
    )
    parser.add_argument(
        "--conviqt-pxx",
        required=False,
        action="store_false",
        help="Beams are in Pxx frame, not Dxx",
        dest="conviqt_dxx",
    )
    parser.add_argument(
        "--conviqt-dxx",
        required=False,
        action="store_true",
        help="Beams are in Dxx frame, not Pxx",
        dest="conviqt_dxx",
    )
    parser.set_defaults(conviqt_dxx=True)
    parser.add_argument(
        "--conviqt-order", default=11, type=np.int, help="Iteration order"
    )
    parser.add_argument(
        "--conviqt-normalize-beam",
        required=False,
        action="store_true",
        help="Normalize the beams",
        dest="conviqt_normalize_beam",
    )
    parser.add_argument(
        "--no-conviqt-normalize-beam",
        required=False,
        action="store_false",
        help="Do not normalize the beams",
        dest="conviqt_normalize_beam",
    )
    parser.set_defaults(conviqt_normalize_beam=False)

    parser.add_argument(
        "--conviqt-calibrate",
        required=False,
        action="store_true",
        help="Normalize the beams",
        dest="conviqt_calibrate",
    )
    parser.add_argument(
        "--no-conviqt-calibrate",
        required=False,
        action="store_false",
        help="Do not normalize the beams",
        dest="conviqt_calibrate",
    )
    parser.set_defaults(conviqt_calibrate=True)
    parser.add_argument(
        "--conviqt-remove-monopole",
        required=False,
        action="store_true",
        help="Remove the sky monopole before convolution",
        dest="conviqt_remove_monopole",
    )
    parser.add_argument(
        "--no-conviqt-remove-monopole",
        required=False,
        action="store_false",
        help="Do not remove the sky monopole before convolution",
        dest="conviqt_remove_monopole",
    )
    parser.set_defaults(conviqt_remove_monopole=False)
    parser.add_argument(
        "--conviqt-remove-dipole",
        required=False,
        action="store_true",
        help="Remove the sky dipole before convolution",
    )
    parser.add_argument(
        "--no-conviqt-remove-dipole",
        required=False,
        action="store_false",
        help="Do not remove the sky dipole before convolution",
    )
    parser.set_defaults(conviqt_remove_dipole=False)
    parser.add_argument(
        "--conviqt-mpi-comm",
        required=False,
        help="MPI communicator used by the OpSimConviqt operator, "
        "either 'rank' or 'group'",
        dest="conviqt_mpi_comm",
    )
    parser.set_defaults(conviqt_mpi_comm="rank")

    add_sky_signal_args(parser)

    return


@function_timer
def apply_conviqt(args, comm, data, cache_prefix="signal", verbose=True, mc=0):
    if (
        args.conviqt_sky_file is None
        or args.conviqt_beam_file is None
        or not args.simulate_sky
    ):
        return None

    log = Logger.get()
    timer = Timer()
    timer.start()

    if comm.world_rank == 0 and verbose:
        log.info("Running Conviqt")

    verbosity = 0
    if verbose:
        verbosity = 1
    if args.debug:
        verbosity = 10

    conviqt = OpSimConviqt(
        getattr(comm, "comm_" + args.conviqt_mpi_comm),
        args.conviqt_sky_file,
        args.conviqt_beam_file,
        lmax=args.conviqt_lmax,
        beammmax=args.conviqt_beam_mmax,
        pol=True,
        fwhm=args.conviqt_fwhm,
        order=args.conviqt_order,
        calibrate=args.conviqt_calibrate,
        dxx=args.conviqt_dxx,
        out=cache_prefix,
        remove_monopole=args.conviqt_remove_monopole,
        remove_dipole=args.conviqt_remove_dipole,
        normalize_beam=args.conviqt_normalize_beam,
        verbosity=verbosity,
        mc=mc,
    )
    conviqt.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Read and sample map")

    return cache_prefix


@function_timer
def apply_weighted_conviqt(args, comm, data, cache_prefix="signal", verbose=True):
    if (
        args.conviqt_sky_file is None
        or args.conviqt_beam_file is None
        or not args.simulate_sky
    ):
        return None

    log = Logger.get()
    timer = Timer()
    timer.start()

    if comm.world_rank == 0 and verbose:
        log.info("Running Weighted Conviqt")

    verbosity = 0
    if verbose:
        verbosity = 1
    if args.debug:
        verbosity = 10

    conviqt = OpSimWeightedConviqt(
        getattr(comm, "comm_" + args.conviqt_mpi_comm),
        args.conviqt_sky_file,
        args.conviqt_beam_file,
        lmax=args.conviqt_lmax,
        beammmax=args.conviqt_beam_mmax,
        pol=True,
        fwhm=args.conviqt_fwhm,
        order=args.conviqt_order,
        calibrate=args.conviqt_calibrate,
        dxx=args.conviqt_dxx,
        out=cache_prefix,
        remove_monopole=args.conviqt_remove_monopole,
        remove_dipole=args.conviqt_remove_dipole,
        normalize_beam=args.conviqt_normalize_beam,
        verbosity=verbosity,
    )
    conviqt.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Read and sample map")

    return cache_prefix


@function_timer
def scan_sky_signal(
    args, comm, data, cache_prefix="signal", verbose=True, pixels="pixels", nnz=3, mc=0
):
    """Scan sky signal from a map."""
    if not args.input_map or not args.simulate_sky:
        return None

    log = Logger.get()
    timer = Timer()
    timer.start()

    # Scan the sky signal

    scansim = OpSimScan(input_map=args.input_map, out=cache_prefix, nnz=nnz, mc=mc)
    scansim.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Read and sample map")

    return cache_prefix


@function_timer
def simulate_sky_signal(
    args, comm, data, focalplanes, cache_prefix, verbose=False, pixels="pixels", mc=0
):
    """Use PySM to simulate smoothed sky signal."""
    if not args.pysm_model or not args.simulate_sky:
        return None
    timer = Timer()
    timer.start()
    fn_cmb = args.pysm_precomputed_cmb_K_CMB
    if fn_cmb is not None:
        fn_cmb = fn_cmb.format(mc)
    # Convolve a signal TOD from PySM
    op_sim_pysm = OpSimPySM(
        data,
        comm=getattr(comm, "comm_" + args.pysm_mpi_comm),
        out=cache_prefix,
        pysm_model=args.pysm_model.split(","),
        pysm_precomputed_cmb_K_CMB=fn_cmb,
        focalplanes=focalplanes,
        apply_beam=args.pysm_apply_beam,
        coord=args.coord,
        pixels=pixels,
    )
    op_sim_pysm.exec(data)
    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0 and verbose:
        timer.report_clear("PySM")

    return cache_prefix



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# FIXME:  This code is deprecated in favor of the load / export model of spt3g data
# reading / writing, as can be found for example in this git repo:
#
#  https://github.com/simonsobs/sotodlib
#
# After identifying the general pieces of those tools, that code should be migrated here
# or to a new git repo focused on TOAST / SPT3G interfacing.


#
# from ..mpi import MPI
#
# import sys
# import os
# import re
#
# import traceback
#
# import numpy as np
#
# from .. import qarray as qa
#
# from ..dist import Data, distribute_discrete
# from ..op import Operator
#
# from .tod import TOD
# from .interval import Interval, intervals_to_chunklist
#
# available = True
# try:
#     from spt3g import core as c3g
#
#     # from spt3g import coordinateutils as c3c
#     from . import spt3g_utils as s3utils
# except:
#     available = False
#
#
# # Module-level constants
#
# STR_TIME = "timestamps"
# STR_FLAG = "flag"
# STR_COMMON = "common"
# STR_QUAT = "quaternion"
# STR_BORE = "boresight"
# STR_BOREAZEL = "boresight_azel"
# STR_POS = "position"
# STR_VEL = "velocity"
# STR_DET = "det"
#
# # Used to determine file breaks when writing new data.  This
# # is the desired approximate file size in bytes.
# TARGET_FRAMEFILE_SIZE = 500000000
#
#
# def read_spt3g_obs(file):
#     reader = c3g.G3Reader(file, 1)
#     f = list(reader(None))
#     obframe = f[0]
#     obspat = re.compile("^obs_(.*)")
#     detpat = re.compile("{}_(.*)".format(STR_QUAT))
#
#     obs = dict()
#     props = dict()
#     dets = dict()
#     nsamp = None
#     for k in obframe.keys():
#         if k == "samples":
#             nsamp = int(obframe[k])
#         else:
#             obsmat = obspat.match(k)
#             detmat = detpat.match(k)
#             if obsmat is not None:
#                 obs[obsmat.group(1)] = s3utils.from_g3_type(obframe[k])
#             elif detmat is not None:
#                 dets[detmat.group(1)] = np.array(obframe[k], dtype=np.float64)
#             else:
#                 props[k] = s3utils.from_g3_type(obframe[k])
#     return obs, props, dets, nsamp
#
#
# def write_spt3g_obs(writer, props, dets, nsamp):
#     f = c3g.G3Frame(c3g.G3FrameType.Observation)
#     for k, v in props.items():
#         f[k] = s3utils.to_g3_type(v)
#     f["samples"] = c3g.G3Int(nsamp)
#     for k, v in dets.items():
#         f["{}_{}".format(STR_QUAT, k)] = c3g.G3VectorDouble(v)
#     writer(f)
#     return
#
#
# def read_spt3g_framesizes(file):
#     sizes = list()
#     obs = True
#     for frame in c3g.G3File(file):
#         if obs:
#             obs = False
#             continue
#         sizes.append(len(frame[STR_TIME]))
#     return sizes
#
#
# def bytes_per_sample(ndet, have_azel):
#     # For each sample we have:
#     #   - 1 x 8 bytes for timestamp
#     #   - 1 x 1 bytes for common flag
#     #   - 4 x 8 bytes for boresight RA/DEC quats
#     #   - 4 x 8 bytes for boresight Az/El quats (optional)
#     #   - 3 x 8 bytes for telescope position
#     #   - 3 x 8 bytes for telescope velocity
#     #   - 1 x 8 bytes x number of dets
#     #   - 1 x 1 bytes x number of dets
#     azbytes = 0
#     if have_azel:
#         azbytes = 32
#     persample = 8 + 1 + 32 + azbytes + 24 + 24 + 9 * ndet
#     return persample
#
#
# class TOD3G(TOD):
#     """This class provides an interface to a directory of spt3g frame files.
#
#     This class serves as *an example* for how to create a TOD class that
#     interfaces with spt3g data files.  Experiments will surely need to
#     implement their own classes that use some of the same utilities and also
#     use their own "schema" for the frames.  This class just reads / writes
#     a minimal set of data required by a TOAST TOD.
#
#     An instance of this class loads a directory of frame files into memory.
#     This data may be manipulated in memory (using the write* methods), but
#     the modifications will not be stored to disk until the export method is
#     called.  If creating a new data set, additional parameters are used to
#     set up the initial detector information.
#
#     Observation properties and detector pointing offsets from the boresight are
#     given as quaternions, and are stored in an observation frame at the
#     start of each frame file.  The sequence of frame files are named with a
#     prefix and a number which represents the first sample of that frame
#     relative to the start of the observation.
#
#     Args:
#         mpicomm (mpi4py.MPI.Comm): the MPI communicator over which this
#             observation data is distributed.
#         detranks (int):  The dimension of the process grid in the detector
#             direction.  The MPI communicator size must be evenly divisible
#             by this number.
#         path (str):  The path to this observation directory.
#         prefix (str):  The prefix string for the frame files.  Different
#             prefix values allow grouping sets of frames for different
#             frequencies, etc in the same observation directory.
#         detectors (dictionary):  (Only when creating new groups) Specify the
#             detector offsets.  Each key is the detector name, and each value is
#             a quaternion tuple.
#         samples (int):  (Only when creating) The number of samples
#             to use.
#         framesizes (list):  (Only when creating) The list of frame sizes.
#             Total must equal samples.
#         azel (bool):  (Only when creating) If True, this TOD will
#             have ground-based data.
#         meta (dict):  (Only when creating) Extra scalar key / value
#             parameters to write.
#         units (G3TimestreamUnits):  (Only when creating new groups) The units
#             of the detector timestreams.
#         detbreaks (list):  Optional list of hard breaks in the detector
#             distribution.
#
#
#     """
#
#     def __init__(
#         self,
#         mpicomm,
#         detranks,
#         path=None,
#         prefix=None,
#         detectors=None,
#         samples=None,
#         framesizes=None,
#         azel=False,
#         meta=dict(),
#         units=None,
#         detbreaks=None,
#     ):
#
#         if not available:
#             raise RuntimeError("spt3g is not available")
#
#         self._path = path
#         self._prefix = prefix
#
#         # When constructing an instance of this class, we are either creating
#         # a new data set in memory or reading from disk.  Make sure that all
#         # the required inputs are given in either case.
#         dets = None
#         nsamp = None
#         props = None
#         self._have_azel = False
#         self._units = None
#         self._frame_sizes = None
#         self._frame_sample_offs = None
#         self._files = None
#         self._file_sample_offs = None
#         self._file_frame_offs = None
#
#         self._createmode = False
#         if (detectors is not None) and (samples is not None):
#             self._createmode = True
#         else:
#             if (self._path is None) or (self._prefix is None):
#                 raise RuntimeError(
#                     "If reading existing data, path and " "prefix are required."
#                 )
#
#         if mpicomm.rank == 0:
#             if self._createmode:
#                 # We are creating a new data set.
#                 dets = detectors
#                 nsamp = samples
#                 props = meta
#                 self._have_azel = azel
#                 self._units = units
#                 self._frame_sizes = framesizes
#                 if self._frame_sizes is None:
#                     self._frame_sizes = [nsamp]
#
#                 # Compute the future frame file breaks that would be used
#                 # if exporting the data.  We ignore the observation frame since
#                 # it is small.
#                 sampbytes = bytes_per_sample(len(dets), azel)
#
#                 (
#                     self._file_sample_offs,
#                     self._file_frame_offs,
#                     self._frame_sample_offs,
#                 ) = s3utils.compute_file_frames(
#                     sampbytes, self._frame_sizes, file_size=TARGET_FRAMEFILE_SIZE
#                 )
#             else:
#                 # We must have existing data
#                 self._frame_sizes = list()
#                 self._frame_sample_offs = list()
#                 self._files = list()
#                 self._file_sample_offs = list()
#                 self._file_frame_offs = list()
#
#                 # Parse frame files to determine sample and frame breaks
#                 fpat = re.compile(r"{}_(\d+).g3".format(prefix))
#
#                 frameoff = 0
#                 checkoff = 0
#                 for root, dirs, files in os.walk(path, topdown=True):
#                     for f in sorted(files):
#                         fmat = fpat.match(f)
#                         if fmat is not None:
#                             ffile = os.path.join(path, f)
#                             # Is this the first frame file?  If so parse the
#                             # observation frame.
#                             if nsamp is None:
#                                 obs, props, dets, nsamp = read_spt3g_obs(ffile)
#                                 props.update(obs)
#                                 if "units" in props:
#                                     self._units = props["units"]
#                                 if "have_azel" in props:
#                                     self._have_azel = props["have_azel"]
#
#                             fsampoff = int(fmat.group(1))
#                             if fsampoff != checkoff:
#                                 raise RuntimeError(
#                                     "frame file {} is at \
#                                     sample offset {}, are some files\
#                                     missing?".format(
#                                         ffile, checkoff
#                                     )
#                                 )
#
#                             self._files.append(ffile)
#                             self._file_sample_offs.append(fsampoff)
#                             self._file_frame_offs.append(frameoff)
#
#                             fsizes = read_spt3g_framesizes(ffile)
#                             fsoff = fsampoff
#                             for fs in fsizes:
#                                 self._frame_sample_offs.append(fsoff)
#                                 fsoff += fs
#                             self._frame_sizes.extend(fsizes)
#                             frameoff += len(fsizes)
#                             checkoff += np.sum(fsizes)
#                     break
#                 if len(self._files) == 0:
#                     raise RuntimeError(
#                         "No frames found at '{}' with prefix '{}'".format(path, prefix)
#                     )
#
#                 # print(self._files)
#                 # print(self._file_frame_offs)
#                 # print(self._file_sample_offs)
#                 # print(self._frame_sizes)
#                 # print(self._frame_sample_offs)
#
#                 # check that the total samples match the obs frame
#                 if nsamp != checkoff:
#                     raise RuntimeError(
#                         "observation frame specifies {}\
#                         samples, but sum of frame files is {}\
#                         samples".format(
#                             nsamp, checkoff
#                         )
#                     )
#
#                 self._file_sample_offs = np.array(
#                     self._file_sample_offs, dtype=np.int64
#                 )
#                 self._file_frame_offs = np.array(self._file_frame_offs, dtype=np.int64)
#                 self._frame_sample_offs = np.array(
#                     self._frame_sample_offs, dtype=np.int64
#                 )
#
#         dets = mpicomm.bcast(dets, root=0)
#         self._detquats = dets
#         nsamp = mpicomm.bcast(nsamp, root=0)
#         props = mpicomm.bcast(props, root=0)
#         self._have_azel = mpicomm.bcast(self._have_azel, root=0)
#         self._units = mpicomm.bcast(self._units, root=0)
#         self._files = mpicomm.bcast(self._files, root=0)
#         self._file_sample_offs = mpicomm.bcast(self._file_sample_offs, root=0)
#         self._file_frame_offs = mpicomm.bcast(self._file_frame_offs, root=0)
#         self._frame_sizes = mpicomm.bcast(self._frame_sizes, root=0)
#         self._frame_sample_offs = mpicomm.bcast(self._frame_sample_offs, root=0)
#
#         # We need to assign a unique integer index to each detector.  This
#         # is used when seeding the streamed RNG in order to simulate
#         # timestreams.  For simplicity, and assuming that detector names
#         # are not too long, we can convert the detector name to bytes and
#         # then to an integer.
#
#         self._detindx = {}
#         for det in dets:
#             bdet = det.encode("utf-8")
#             uid = None
#             try:
#                 ind = int.from_bytes(bdet, byteorder="little")
#                 uid = int(ind & 0xFFFFFFFF)
#             except:
#                 raise RuntimeError(
#                     "Cannot convert detector name {} to a unique integer-\
#                     maybe it is too long?".format(
#                         det
#                     )
#                 )
#             self._detindx[det] = uid
#
#         # call base class constructor to distribute data
#         super().__init__(
#             mpicomm,
#             list(sorted(dets.keys())),
#             nsamp,
#             detindx=self._detindx,
#             detranks=detranks,
#             detbreaks=detbreaks,
#             sampsizes=self._frame_sizes,
#             meta=props,
#         )
#
#         # Now that the data distribution is set, read frames into cache
#         # if needed.
#
#         self._done_cache_init = False
#
#         if not self._createmode:
#             # We are reading existing data into cache.  Create our local pieces
#             # of the data.
#             self.load_frames()
#         return
#
#     def _cache_init(self):
#         if not self._done_cache_init:
#             # Timestamps
#             self.cache.create(STR_TIME, np.int64, (self.local_samples[1],))
#
#             # Boresight quaternions
#             self.cache.create(STR_BORE, np.float64, (self.local_samples[1], 4))
#             if self._have_azel:
#                 self.cache.create(STR_BOREAZEL, np.float64, (self.local_samples[1], 4))
#
#             # Common flags
#             name = "{}_{}".format(STR_FLAG, STR_COMMON)
#             self.cache.create(name, np.uint8, (self.local_samples[1],))
#
#             # Telescope position and velocity
#             self.cache.create(STR_POS, np.float64, (self.local_samples[1], 3))
#             self.cache.create(STR_VEL, np.float64, (self.local_samples[1], 3))
#
#             # Detector data and flags
#             for det in self.local_dets:
#                 name = "{}_{}".format(STR_DET, det)
#                 self.cache.create(name, np.float64, (self.local_samples[1],))
#                 name = "{}_{}".format(STR_FLAG, det)
#                 self.cache.create(name, np.uint8, (self.local_samples[1],))
#
#             self._done_cache_init = True
#
#         return
#
#     def load_frames(self):
#         self._cache_init()
#         for ifile, (ffile, foff) in enumerate(zip(self._files, self._file_frame_offs)):
#             nframes = None
#             if ifile == len(self._files) - 1:
#                 # we are at the last file
#                 nframes = len(self._frame_sizes) - foff
#             else:
#                 # get number of frames in this file
#                 nframes = self._file_frame_offs[ifile + 1] - foff
#
#             # print("load {}:  proc {} file {} starts at frame {} and has {} frames".format(self._path, self.mpicomm.rank, ifile, foff, nframes), flush=True)
#
#             # nframes includes only the scan frames.  We add one here to
#             # get the total including the observation frame.
#             gfile = [None for x in range(nframes + 1)]
#             if self.mpicomm.rank == 0:
#                 gfile = c3g.G3File(ffile)
#
#             obs = True
#             for fileframe, fdata in enumerate(gfile):
#                 if obs:
#                     # Skip the observation frame.
#                     obs = False
#                     continue
#                 scanframe = fileframe - 1
#                 frame = foff + scanframe
#                 frame_offset = self._frame_sample_offs[frame]
#                 frame_size = self._frame_sizes[frame]
#                 # if self.mpicomm.rank == 0:
#                 #     print("load {}:    proc {} working on frame {} (samples {} - {}) data = {}".format(self._path, self.mpicomm.rank, frame, frame_offset, frame_offset+frame_size, fdata), flush=True)
#
#                 s3utils.frame_to_cache(
#                     self,
#                     frame,
#                     frame_offset,
#                     frame_size,
#                     frame_data=fdata,
#                     detector_map=STR_DET,
#                     flag_map=STR_FLAG,
#                     common_prefix=None,
#                     det_prefix="{}_".format(STR_DET),
#                     flag_prefix="{}_".format(STR_FLAG),
#                 )
#
#                 # print("load {}:    proc {} hit barrier for frame {}".format(self._path, self.mpicomm.rank, frame), flush=True)
#                 self.mpicomm.barrier()
#                 # if self.mpicomm.rank == 0:
#                 # print("load {}:    proc {} finished frame {}".format(self._path, self.mpicomm.rank, frame), flush=True)
#             del gfile
#             # print("load {}:    proc {} finished file {}".format(self._path, self.mpicomm.rank, ifile), flush=True)
#         return
#
#     def _export_frames(
#         self,
#         path=None,
#         prefix=None,
#         cache_name=None,
#         cache_common=None,
#         cache_flag_name=None,
#     ):
#         """Export cache data to frames.
#
#         This will either export the cache fields that correspond to the "real"
#         data (those manipulated by the read / write methods) or will export
#         alternate cache fields specified by the arguments.
#
#         Args:
#             path (str):  Override any path specified at construction time.
#             prefix (str):  Override any prefix specified at construction time.
#             cache_name (str):  When exporting data, the name of the cache
#                 object (<name>_<detector>) to use for the detector timestream.
#                 If None, use the TOD read* methods.
#             cache_common (str):  When exporting data, the name of the cache
#                 object to use for common flags.  If None, use the TOD read*
#                 methods.
#             cache_flag_name (str):  When exporting data, the name of the
#                 cache object (<name>_<detector>) to use for the detector
#                 flags.  If None, use the TOD read* methods.
#
#         """
#         self._cache_init()
#
#         # Create the frame schema we are using when exporting data
#
#         common_fields = [
#             (STR_TIME, c3g.G3VectorTime, STR_TIME),
#             (STR_BORE, c3g.G3VectorDouble, STR_BORE),
#             (STR_BOREAZEL, c3g.G3VectorDouble, STR_BOREAZEL),
#             (STR_POS, c3g.G3VectorDouble, STR_POS),
#             (STR_VEL, c3g.G3VectorDouble, STR_VEL),
#         ]
#         if cache_common is None:
#             cname = "{}_{}".format(STR_FLAG, STR_COMMON)
#             common_fields.append((cname, c3g.G3VectorUnsignedChar, cname))
#         else:
#             common_fields.append((cache_common, c3g.G3VectorUnsignedChar, cache_common))
#
#         det_fields = None
#         if cache_name is None:
#             det_fields = [("{}_{}".format(STR_DET, d), d) for d in self.detectors]
#         else:
#             det_fields = [("{}_{}".format(cache_name, d), d) for d in self.detectors]
#
#         flag_fields = None
#         if cache_flag_name is None:
#             flag_fields = [("{}_{}".format(STR_FLAG, d), d) for d in self.detectors]
#         else:
#             flag_fields = [
#                 ("{}_{}".format(cache_flag_name, d), d) for d in self.detectors
#             ]
#
#         ex_path = self._path
#         if path is not None:
#             ex_path = path
#
#         ex_prefix = self._prefix
#         if prefix is not None:
#             ex_prefix = prefix
#
#         if (ex_path is None) or (ex_prefix is None):
#             raise RuntimeError(
#                 "You must specify the TOD path and prefix, "
#                 "either at construction or export"
#             )
#
#         ex_files = [
#             os.path.join(ex_path, "{}_{:08d}.g3".format(ex_prefix, x))
#             for x in self._file_sample_offs
#         ]
#
#         for ifile, (ffile, foff) in enumerate(zip(ex_files, self._file_frame_offs)):
#             nframes = None
#             # print("  ifile = {}, ffile = {}, foff = {}".format(ifile, ffile, foff), flush=True)
#             if ifile == len(ex_files) - 1:
#                 # we are at the last file
#                 nframes = len(self._frame_sizes) - foff
#             else:
#                 # get number of frames in this file
#                 nframes = self._file_frame_offs[ifile + 1] - foff
#
#             writer = None
#             if self.mpicomm.rank == 0:
#                 writer = c3g.G3Writer(ffile)
#                 props = self.meta()
#                 props["units"] = self._units
#                 props["have_azel"] = self._have_azel
#                 # print("Writing props:")
#                 # print(props, flush=True)
#                 write_spt3g_obs(writer, props, self._detquats, self.total_samples)
#
#             # Collect data for all frames in the file in one go.
#
#             frm_offsets = [self._frame_sample_offs[foff + f] for f in range(nframes)]
#             frm_sizes = [self._frame_sizes[foff + f] for f in range(nframes)]
#
#             # if self.mpicomm.rank == 0:
#             #     print("  {} file {}".format(self._path, ifile), flush=True)
#             #     print("    start frame = {}, nframes = {}".format(foff, nframes), flush=True)
#             #     print("    frame offs = ",frm_offsets, flush=True)
#             #     print("    frame sizes = ",frm_sizes, flush=True)
#
#             fdata = s3utils.cache_to_frames(
#                 self,
#                 foff,
#                 nframes,
#                 frm_offsets,
#                 frm_sizes,
#                 common=common_fields,
#                 detector_fields=det_fields,
#                 flag_fields=flag_fields,
#                 detector_map=STR_DET,
#                 flag_map=STR_FLAG,
#                 units=self._units,
#             )
#
#             if self.mpicomm.rank == 0:
#                 for fdt in fdata:
#                     writer(fdt)
#                 del writer
#             del fdata
#
#         return
#
#     def detoffset(self):
#         return dict(self._detquats)
#
#     def _get_boresight(self, start, n):
#         self._cache_init()
#         ref = self.cache.reference(STR_BORE)[start : start + n, :]
#         return ref
#
#     def _put_boresight(self, start, data):
#         self._cache_init()
#         ref = self.cache.reference(STR_BORE)
#         ref[start : (start + data.shape[0]), :] = data
#         del ref
#         return
#
#     def _get_boresight_azel(self, start, n):
#         if not self._have_azel:
#             raise RuntimeError("No Az/El pointing for this TOD")
#         self._cache_init()
#         ref = self.cache.reference(STR_BOREAZEL)[start : start + n, :]
#         return ref
#
#     def _put_boresight_azel(self, start, data):
#         if not self._have_azel:
#             raise RuntimeError("No Az/El pointing for this TOD")
#         self._cache_init()
#         ref = self.cache.reference(STR_BOREAZEL)
#         ref[start : (start + data.shape[0]), :] = data
#         del ref
#         return
#
#     def _get(self, detector, start, n):
#         self._cache_init()
#         name = "{}_{}".format(STR_DET, detector)
#         ref = self.cache.reference(name)[start : start + n]
#         return ref
#
#     def _put(self, detector, start, data):
#         self._cache_init()
#         name = "{}_{}".format(STR_DET, detector)
#         ref = self.cache.reference(name)
#         ref[start : (start + data.shape[0])] = data
#         del ref
#         return
#
#     def _get_flags(self, detector, start, n):
#         self._cache_init()
#         name = "{}_{}".format(STR_FLAG, detector)
#         ref = self.cache.reference(name)[start : start + n]
#         return ref
#
#     def _put_flags(self, detector, start, flags):
#         self._cache_init()
#         name = "{}_{}".format(STR_FLAG, detector)
#         ref = self.cache.reference(name)
#         ref[start : (start + flags.shape[0])] = flags
#         del ref
#         return
#
#     def _get_common_flags(self, start, n):
#         self._cache_init()
#         name = "{}_{}".format(STR_FLAG, STR_COMMON)
#         ref = self.cache.reference(name)[start : start + n]
#         return ref
#
#     def _put_common_flags(self, start, flags):
#         self._cache_init()
#         name = "{}_{}".format(STR_FLAG, STR_COMMON)
#         ref = self.cache.reference(name)
#         ref[start : (start + flags.shape[0])] = flags
#         del ref
#         return
#
#     def _get_times(self, start, n):
#         self._cache_init()
#         ref = self.cache.reference(STR_TIME)[start : start + n]
#         tm = 1.0e-9 * ref.astype(np.float64)
#         del ref
#         return tm
#
#     def _put_times(self, start, stamps):
#         self._cache_init()
#         ref = self.cache.reference(STR_TIME)
#         ref[start : (start + stamps.shape[0])] = np.array(
#             1.0e9 * stamps, dtype=np.int64
#         )
#         del ref
#         return
#
#     def _get_pntg(self, detector, start, n):
#         self._cache_init()
#         # Get boresight pointing (from disk or cache)
#         bore = self._get_boresight(start, n)
#         # Apply detector quaternion and return
#         return qa.mult(bore, self._detquats[detector])
#
#     def _put_pntg(self, detector, start, data):
#         raise RuntimeError(
#             "TOD3G computes detector pointing on the fly."
#             " Use the write_boresight() method instead."
#         )
#         return
#
#     def _get_position(self, start, n):
#         self._cache_init()
#         ref = self.cache.reference(STR_POS)[start : start + n, :]
#         return ref
#
#     def _put_position(self, start, pos):
#         self._cache_init()
#         ref = self.cache.reference(STR_POS)
#         ref[start : (start + pos.shape[0]), :] = pos
#         del ref
#         return
#
#     def _get_velocity(self, start, n):
#         self._cache_init()
#         ref = self.cache.reference(STR_VEL)[start : start + n, :]
#         return ref
#
#     def _put_velocity(self, start, vel):
#         self._cache_init()
#         ref = self.cache.reference(STR_VEL)
#         ref[start : (start + vel.shape[0]), :] = vel
#         del ref
#         return
#
#     def export(self, **kwargs):
#         self._export_frames(**kwargs)
#         return
#
#
# def obsweight_spt3g(framefile):
#     """Returns a weight for a framefile.
#
#     This reads the special observation frame at the start of the file and
#     returns the product of the number of detectors and the number of
#     samples.
#
#     Args:
#         framefile (str):  The frame file.  The first frame should be an
#             observation frame.
#
#     Returns:
#         (float): the weight.
#
#     """
#     obs, props, dets, nsamp = read_spt3g_obs(framefile)
#     return nsamp * len(dets.keys())
#
#
# def load_spt3g(comm, detranks, path, prefix, obsweight, todclass, **kwargs):
#     """
#     Loads an existing TOAST dataset in SPT3G format.
#
#     This takes a 2-level TOAST communicator and opens existing spt3g
#     directories of frame files.  For each observation (directory of frame
#     files), the specified TOD class is instantiated.
#
#     Args:
#         comm (toast.Comm): the toast Comm class for distributing the data.
#         detranks (int):  The dimension of the process grid in the detector
#             direction.  The MPI communicator size must be evenly divisible
#             by this number.
#         path (str):  The top-level directory that contains subdirectories (one
#             per observation).
#         prefix (str):  The frame file prefix.
#         obsweight (function):  A function that returns the relative weight of
#             an observation, given the first frame file of that observation.
#             This will only be called on the root process.
#         todclass (TOD): an SPT3G-compatible TOD class, which must have a
#             constructor that takes the communicator, detranks, path and
#             prefix.  All additional arguments should be keyword args.
#         kwargs: All additional arguments to this function are passed to the
#             TOD constructor for each observation.
#
#     Returns (toast.Data):
#         The distributed data object.
#
#     """
#     if not available:
#         raise RuntimeError("spt3g is not available")
#         return None
#     # the global communicator
#     cworld = comm.comm_world
#     # the communicator within the group
#     cgroup = comm.comm_group
#     # the communicator with all processes with
#     # the same rank within their group
#     crank = comm.comm_rank
#
#     # One process gets the list of observation directories
#     obslist = list()
#     weight = dict()
#
#     if cworld.rank == 0:
#         for root, dirs, files in os.walk(path, topdown=True):
#             for d in dirs:
#                 # FIXME:  Add some check here to make sure that this is a
#                 # directory of frame files.
#                 obslist.append(d)
#                 # Read the observation frame to find the number of samples.
#                 fr = os.path.join(path, d, "{}_{:08d}.g3".format(prefix, 0))
#                 weight[d] = obsweight(fr)
#             break
#         obslist = sorted(obslist)
#
#     obslist = cworld.bcast(obslist, root=0)
#     weight = cworld.bcast(weight, root=0)
#
#     # Distribute observations based on number of samples
#     dweight = [weight[x] for x in obslist]
#     distobs = distribute_discrete(dweight, comm.ngroups)
#
#     # Distributed data
#
#     data = Data(comm)
#
#     # Now every group adds its observations to the list
#
#     firstobs = distobs[comm.group][0]
#     nobs = distobs[comm.group][1]
#     for ob in range(firstobs, firstobs + nobs):
#         opath = os.path.join(path, obslist[ob])
#         fr = os.path.join(opath, "{}_{:08d}.g3".format(prefix, 0))
#         obs, props, dets, nsamp = read_spt3g_obs(fr)
#         # if cgroup.rank == 0:
#         #     print("group {} creating TOD {}".format(comm.group, opath), flush=True)
#         try:
#             obs["tod"] = todclass(cgroup, detranks, path=opath, prefix=prefix, **kwargs)
#         except:
#             exc_type, exc_value, exc_traceback = sys.exc_info()
#             lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
#             lines = ["Proc {}: {}".format(MPI.COMM_WORLD.rank, x) for x in lines]
#             print("".join(lines), flush=True)
#             MPI.COMM_WORLD.Abort()
#         # print("proc {} hit TOD {} barrier".format(comm.comm_world.rank, opath), flush=True)
#         comm.comm_group.barrier()
#         data.obs.append(obs)
#
#     return data
#
#
# class Op3GExport(Operator):
#     """Operator which writes data to a directory tree of frame files.
#
#     The top level directory will contain one subdirectory per observation.
#     Each observation directory will contain one frame file per sample chunk.
#
#     Args:
#         outdir (str): the top-level output directory.
#         todclass (TOD): a SPT3G-compatible TOD class, which must have a
#             constructor with the proper required arguments and which takes
#             additional arguments as keywords.
#         use_todchunks (bool): if True, use the chunks of the original TOD for
#             data distribution.
#         use_intervals (bool): if True, use the intervals in the observation
#             dictionary for data distribution.
#         ctor_opts (dict): dictionary of options to pass as kwargs to the
#             TOD constructor.
#         export_opts (dict): dictionary of options to pass to the export method
#             of the TOD.
#         cache_name (str):  The name of the cache object (<name>_<detector>) in
#             the existing TOD to use for the detector timestream.  If None, use
#             the read* methods from the existing TOD.
#         cache_common (str):  The name of the cache object in the existing TOD
#             to use for common flags.  If None, use the read* methods from the
#             existing TOD.
#         cache_flag_name (str):   The name of the cache object
#             (<name>_<detector>) in the existing TOD to use for the flag
#             timestream.  If None, use the read* methods from the existing TOD.
#
#     """
#
#     def __init__(
#         self,
#         outdir,
#         todclass,
#         use_todchunks=False,
#         use_intervals=False,
#         ctor_opts={},
#         export_opts={},
#         cache_common=None,
#         cache_name=None,
#         cache_flag_name=None,
#     ):
#
#         if not available:
#             raise RuntimeError("spt3g is not available")
#
#         self._outdir = outdir
#         self._todclass = todclass
#         self._ctor_opts = ctor_opts
#         self._export_opts = export_opts
#         self._cache_common = cache_common
#         self._cache_name = cache_name
#         self._cache_flag_name = cache_flag_name
#         if use_todchunks and use_intervals:
#             raise RuntimeError("cannot use both TOD chunks and Intervals")
#         self._usechunks = use_todchunks
#         self._useintervals = use_intervals
#         # We call the parent class constructor
#         super().__init__()
#
#     def exec(self, data):
#         """Export data to a directory tree of SPT3G frames.
#
#         For errors that prevent the export, this function will directly call
#         MPI Abort() rather than raise exceptions.  This could be changed in
#         the future if additional logic is implemented to ensure that all
#         processes raise an exception when one process encounters an error.
#
#         Args:
#             data (toast.Data): The distributed data.
#
#         """
#         # the two-level toast communicator
#         comm = data.comm
#         # the global communicator
#         cworld = comm.comm_world
#         # the communicator within the group
#         cgroup = comm.comm_group
#         # the communicator with all processes with
#         # the same rank within their group
#         crank = comm.comm_rank
#
#         # One process checks the path
#         if cworld.rank == 0:
#             if not os.path.isdir(self._outdir):
#                 os.makedirs(self._outdir)
#         cworld.barrier()
#
#         for obs in data.obs:
#             # Observation information.  Anything here that is a simple data
#             # type will get written to the observation frame.
#             props = dict()
#             for k, v in obs.items():
#                 if isinstance(v, (int, str, bool, float)):
#                     props["obs_{}".format(k)] = v
#
#             # Every observation must have a name...
#             obsname = obs["name"]
#
#             # if cworld.rank == 0:
#             #     print("Start spt3g data copy for {}".format(obsname), flush=True)
#
#             # The existing TOD
#             oldtod = obs["tod"]
#             nsamp = oldtod.total_samples
#             dets = oldtod.detoffset()
#
#             # print("rank {}, obs {}:  old tod has {} samps of {}".format(cworld.rank, obsname, nsamp, sorted(dets.keys())), flush=True)
#
#             # Get any other metadata from the old TOD
#             props.update(oldtod.meta())
#
#             # First process in the group makes the output directory
#             obsdir = os.path.join(self._outdir, obsname)
#             if cgroup.rank == 0:
#                 if not os.path.isdir(obsdir):
#                     os.makedirs(obsdir)
#             # print("rank {}, obs {}:  hit group barrier".format(cworld.rank, obsname), flush=True)
#             cgroup.barrier()
#
#             olddetranks, oldsampranks = oldtod.grid_size
#
#             # Do we have ground pointing?
#             azel = None
#             if cgroup.rank == 0:
#                 try:
#                     test = oldtod.read_boresight_azel(local_start=0, n=1)
#                     azel = True
#                     del test
#                 except:
#                     azel = False
#             azel = cgroup.bcast(azel, root=0)
#             # print("rank {}, obs {}:  azel = {}".format(cworld.rank, obsname, azel), flush=True)
#
#             # Determine data distribution chunks
#             framesizes = None
#             if self._usechunks:
#                 framesizes = oldtod.total_chunks
#             elif self._useintervals:
#                 if "intervals" not in obs:
#                     raise RuntimeError(
#                         "Observation does not contain intervals"
#                         ", cannot distribute using them"
#                     )
#                 framesizes = intervals_to_chunklist(obs["intervals"], nsamp)
#
#             # The new TOD to handle exporting
#             tod = self._todclass(
#                 oldtod.mpicomm,
#                 olddetranks,
#                 detectors=dets,
#                 samples=nsamp,
#                 framesizes=framesizes,
#                 azel=azel,
#                 meta=props,
#                 **self._ctor_opts
#             )
#
#             # Copy data between TODs
#
#             # print("rank {}, obs {}:  start data copy".format(cworld.rank, obsname), flush=True)
#
#             tod.write_times(stamps=oldtod.read_times())
#             tod.write_boresight(data=oldtod.read_boresight())
#             if azel:
#                 tod.write_boresight_azel(data=oldtod.read_boresight_azel())
#             tod.write_position(pos=oldtod.read_position())
#             tod.write_velocity(vel=oldtod.read_velocity())
#             if self._cache_common is None:
#                 tod.write_common_flags(flags=oldtod.read_common_flags())
#             else:
#                 ref = oldtod.cache.reference(self._cache_common)
#                 tod.write_common_flags(flags=ref)
#                 del ref
#             for d in oldtod.local_dets:
#                 if self._cache_name is None:
#                     tod.write(detector=d, data=oldtod.read(detector=d))
#                 else:
#                     ref = oldtod.cache.reference("{}_{}".format(self._cache_name, d))
#                     tod.write(detector=d, data=ref)
#                     del ref
#                 if self._cache_flag_name is None:
#                     tod.write_flags(detector=d, flags=oldtod.read_flags(detector=d))
#                 else:
#                     ref = oldtod.cache.reference(
#                         "{}_{}".format(self._cache_flag_name, d)
#                     )
#                     tod.write_flags(detector=d, flags=ref)
#                     del ref
#
#             # print("rank {}, obs {}:  end data copy".format(cworld.rank, obsname), flush=True)
#
#             # if cgroup.rank == 0:
#             #     print("Start spt3g data export for {}".format(obsname), flush=True)
#
#             # Export data from cache.
#             tod.export(path=obsdir, **self._export_opts)
#
#             # if cgroup.rank == 0:
#             #     print("Done spt3g export for {}".format(obsname), flush=True)
#
#         return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import copy

import os

import numpy as np

import scipy.signal

import astropy.io.fits as pf

from .. import qarray as qa

from ..timing import Timer, function_timer

from ..todmap import MapSampler

from ..tod import flagged_running_average, Interval

from .psd_math import autocov_psd, crosscov_psd


class OpNoiseEstim:
    """Noise estimation operator.

    Args:
        signal(str):  Cache object name to analyze
        flags(str):  Cached flags to apply
        detmask(byte):  Flag bits to consider
        commonmask(byte):  Flag bits to consider
        out(str):  Output directory to write the PSDs to.
        maskfile(str):  FITS file name to read the mask from.
        mapfile(str):  FITS map to sample and subtract from the signal.
        pol(bool):  Sample also the polarized part of the map.
        nbin_psd(int):  Bin the resulting PSD.
        lagmax(int):  Maximum lag to consider for the covariance function.
        stationary_period(float):  Break the observation into several
            estimation periods of this length [s].
        nosingle(bool):  Do not evaluate individual PSDs.  Overridden by pairs.
        nocross(bool):  Do not evaluate cross-PSDs.  Overridden by pairs.
        calibrate_signal_estimate(bool):  Regress, not just subtract the
            signal estimate.
        nsum(int):  Downsampling factor for decimated data.
        naverage(int):  Smoothing kernel width for downsampled data.
        apply_intervals(bool):  If true, only measure the covariance
             within each interval
        pairs(iterable):  Detector pairs to estimate noise for.  Overrides
            nosingle and nocross.
        save_cov(bool):  Save also the sample covariance.

    """

    def __init__(
        self,
        signal=None,
        flags=None,
        detmask=1,
        commonmask=1,
        out=None,
        maskfile=None,
        mapfile=None,
        pol=True,
        nbin_psd=1000,
        lagmax=10000,
        stationary_period=86400.0,
        nosingle=False,
        nocross=True,
        calibrate_signal_estimate=False,
        nsum=10,
        naverage=100,
        apply_intervals=False,
        pairs=None,
        save_cov=False,
    ):
        self._signal = signal
        self._flags = flags
        self._detmask = detmask
        self._commonmask = commonmask
        self._out = out
        self._maskfile = maskfile
        self._mapfile = mapfile
        self._pol = pol
        self._nbin_psd = nbin_psd
        self._lagmax = lagmax
        self._stationary_period = stationary_period
        self._nosingle = nosingle
        self._nocross = nocross
        self._calibrate_signal_estimate = calibrate_signal_estimate
        self._apply_intervals = apply_intervals
        self._pairs = pairs
        # Parameters for downsampling the data
        self._nsum = nsum
        self._naverage = naverage
        self._save_cov = save_cov

    @function_timer
    def exec(self, data):
        comm = data.comm.comm_group

        masksampler = None
        if self._maskfile:
            masksampler = MapSampler(self._maskfile, comm=comm)
        mapsampler = None
        if self._mapfile:
            mapsampler = MapSampler(self._mapfile, comm=comm, pol=True)

        for obs in data.obs:
            tod = obs["tod"]
            if len(tod.local_dets) < len(tod.detectors):
                raise RuntimeError(
                    "Noise estimation does not work on " "detector-split data"
                )
            dets = {}
            for idet, det in enumerate(tod.detectors):
                dets[det] = idet
            ndet = len(dets)
            if self._pairs is not None:
                pairs = self._pairs
            else:
                # Construct a list of detector pairs
                pairs = []
                for idet1 in range(ndet):
                    det1 = tod.detectors[idet1]
                    for idet2 in range(idet1, ndet):
                        det2 = tod.detectors[idet2]
                        if det1 == det2 and self._nosingle:
                            continue
                        if det1 != det2 and self._nocross:
                            continue
                        pairs.append([det1, det2])
            timestamps = tod.local_times()
            commonflags = tod.local_common_flags()
            commonflags = commonflags & self._commonmask != 0
            fsample = 1 / np.median(np.diff(timestamps))

            if "name" in obs:
                fileroot = "noise_{}".format(obs["name"])
            elif "id" in obs:
                fileroot = "noise_{}".format(obs["id"])
            else:
                fileroot = "noise_{}".format(int(timestamps[0]))

            if self._apply_intervals:
                intervals = tod.local_intervals(obs["intervals"])
            else:
                intervals = [
                    Interval(
                        start=timestamps[0],
                        stop=timestamps[-1],
                        first=0,
                        last=timestamps.size - 1,
                    )
                ]

            self.subtract_signal(tod, comm, masksampler, mapsampler, intervals)

            # self.highpass_signal(tod, comm, intervals)

            # Extend the gap between intervals to prevent sample pairs
            # that cross the gap.

            gap_min = np.int(self._lagmax) + 1
            # Downsampled data requires longer gaps
            gap_min_nsum = np.int(self._lagmax * self._nsum) + 1
            offset, nsamp = tod.local_samples
            gapflags = np.zeros_like(commonflags)
            gapflags_nsum = np.zeros_like(commonflags)
            for ival1, ival2 in zip(intervals[:-1], intervals[1:]):
                gap_start = ival1.last + 1
                gap_stop = max(gap_start + gap_min, ival2.first)
                gap_stop_nsum = max(gap_start + gap_min_nsum, ival2.first)
                if gap_start < offset + nsamp and gap_stop > offset:
                    gap_start = max(0, gap_start - offset)
                    gap_stop = min(offset + nsamp, gap_stop - offset)
                    gapflags[gap_start:gap_stop] = True
                    gap_stop_nsum = min(offset + nsamp, gap_stop_nsum - offset)
                    gapflags_nsum[gap_start:gap_stop_nsum] = True

            # FIXME: This operator needs to handle situations where
            # det1 and det2 are not on the same process.  Then the
            # check at the top of the loop can be removed.

            for det1, det2 in pairs:
                if det1 not in dets or det2 not in dets:
                    # User-specified pair is invalid
                    continue
                signal1 = tod.local_signal(det1)
                flags1 = tod.local_flags(det1, name=self._flags)
                flags = flags1 & self._detmask != 0
                signal2 = None
                flags2 = None
                if det1 != det2:
                    signal2 = tod.local_signal(det2)
                    flags2 = tod.local_flags(det2, name=self._flags)
                    flags[flags2 & self._detmask != 0] = True
                flags[commonflags] = True

                self.process_noise_estimate(
                    signal1,
                    signal2,
                    flags,
                    gapflags,
                    gapflags_nsum,
                    timestamps,
                    fsample,
                    comm,
                    fileroot,
                    det1,
                    det2,
                    intervals,
                )

        return

    def highpass_signal(self, tod, comm, intervals):
        """Suppress the sub-harmonic modes in the TOD by high-pass
        filtering.
        """
        timer = Timer()
        timer.start()
        rank = 0
        if comm is not None:
            rank = comm.rank
        if rank == 0:
            print("High-pass-filtering signal", flush=True)
        for det in tod.local_dets:
            signal = tod.local_signal(det, name=self._signal)
            flags = tod.local_flags(det, name=self._flags)
            flags &= self._detmask
            for ival in intervals:
                ind = slice(ival.first, ival.last + 1)
                sig = signal[ind]
                flg = flags[ind]
                trend = flagged_running_average(
                    sig, flg, self._lagmax, return_flags=False
                )
                sig -= trend
        if comm is not None:
            comm.barrier()
        if rank == 0:
            timer.report_clear("TOD high pass")
        return

    def subtract_signal(self, tod, comm, masksampler, mapsampler, intervals):
        """Subtract a signal estimate from the TOD and update the
        flags for noise estimation.
        """
        if mapsampler is None and masksampler is None:
            return
        timer = Timer()
        timer.start()
        rank = 0
        if comm is not None:
            rank = comm.rank
        if rank == 0:
            print("Subtracting signal", flush=True)
        for det in tod.local_dets:
            if det.endswith("-diff") and not self._pol:
                continue
            # if comm.rank == 0:
            #    print('Subtracting signal for {}'.format(det), flush=True)
            #    tod.cache.report()
            epsilon = 0  # FIXME: Where can we get this for every detector?
            eta = (1 - epsilon) / (1 + epsilon)
            signal = tod.local_signal(det, name=self._signal)
            flags = tod.local_flags(det, name=self._flags)
            flags &= self._detmask
            try:
                quats = tod.local_pointing(det)
            except AttributeError:
                quats = None
            if quats is None:
                continue
            iquweights = tod.local_weights(det)
            for ival in intervals:
                ind = slice(ival.first, ival.last + 1)
                sig = signal[ind]
                flg = flags[ind]
                quat = quats[ind]
                theta, phi = qa.to_position(quat)
                iquw = iquweights[ind, :]
                if masksampler is not None:
                    maskflg = masksampler.at(theta, phi) < 0.5
                    flg[maskflg] |= 255
                if mapsampler is not None:
                    if self._pol:
                        bg = mapsampler.atpol(theta, phi, iquw)
                    else:
                        bg = mapsampler.at(theta, phi) * iquw[:, 0]
                    if self._calibrate_signal_estimate:
                        good = flg == 0
                        ngood = np.sum(good)
                        if ngood > 1:
                            templates = np.vstack([np.ones(ngood), bg[good]])
                            invcov = np.dot(templates, templates.T)
                            cov = np.linalg.inv(invcov)
                            proj = np.dot(templates, sig[good])
                            coeff = np.dot(cov, proj)
                            bg = coeff[0] + coeff[1] * bg
                    sig -= bg
        if comm is not None:
            comm.barrier()
        if rank == 0:
            timer.report_clear("TOD signal subtraction")
        return

    def decimate(self, x, flg, gapflg, intervals):
        # Low-pass filter with running average, then downsample
        xx = x.copy()
        flags = flg.copy()
        for ival in intervals:
            ind = slice(ival.first, ival.last + 1)
            xx[ind], flags[ind] = flagged_running_average(
                x[ind], flg[ind], self._naverage, return_flags=True
            )
        return xx[:: self._nsum].copy(), (flags + gapflg)[:: self._nsum].copy()

    # def highpass(self, x, flg):
    #     # Flagged real-space high pass filter
    #     xx = x.copy()
    #
    #     j = 0
    #     while j < x.size and flg[j]: j += 1
    #
    #     alpha = .999
    #
    #     for i in range(j+1, x.size):
    #         if flg[i]:
    #             xx[i] = x[j]
    #         else:
    #             xx[i] = alpha*(xx[j] + x[i] - x[j])
    #             j = i
    #
    #     xx /= alpha
    #     return xx

    def log_bin(self, freq, nbin=100, fmin=None, fmax=None):
        if np.any(freq == 0):
            raise Exception("Logarithmic binning should not include " "zero frequency")

        if fmin is None:
            fmin = np.amin(freq)
        if fmax is None:
            fmax = np.amax(freq)

        bins = np.logspace(
            np.log(fmin), np.log(fmax), num=nbin + 1, endpoint=True, base=np.e
        )
        bins[-1] *= 1.01  # Widen the last bin not to have a bin with one entry

        locs = np.digitize(freq, bins).astype(np.int32)
        hits = np.zeros(nbin + 2, dtype=np.int32)
        for loc in locs:
            hits[loc] += 1
        return locs, hits

    def bin_psds(self, my_psds, fmin=None, fmax=None):
        my_binned_psds = []
        my_times = []
        binfreq0 = None

        for i in range(len(my_psds)):
            t0, _, freq, psd = my_psds[i]

            good = freq != 0

            if self._nbin_psd is not None:
                locs, hits = self.log_bin(
                    freq[good], nbin=self._nbin_psd, fmin=fmin, fmax=fmax
                )
                binfreq = np.zeros(hits.size)
                for loc, f in zip(locs, freq[good]):
                    binfreq[loc] += f
                binfreq = binfreq[hits != 0] / hits[hits != 0]
            else:
                binfreq = freq
                hits = np.ones(len(binfreq))

            if binfreq0 is None:
                binfreq0 = binfreq
            else:
                if np.any(binfreq != binfreq0):
                    raise Exception("Binned PSD frequencies change")

            if self._nbin_psd is not None:
                binpsd = np.zeros(hits.size)
                for loc, p in zip(locs, psd[good]):
                    binpsd[loc] += p
                binpsd = binpsd[hits != 0] / hits[hits != 0]
            else:
                binpsd = psd

            my_times.append(t0)
            my_binned_psds.append(binpsd)
        return my_binned_psds, my_times, binfreq0

    def discard_outliers(self, binfreq, all_psds, all_times, all_cov):
        all_psds = copy.deepcopy(all_psds)
        all_times = copy.deepcopy(all_times)
        if self._save_cov:
            all_cov = copy.deepcopy(all_cov)

        nrow, ncol = np.shape(all_psds)

        # Discard empty PSDs

        i = 1
        nempty = 0
        while i < nrow:
            p = all_psds[i]
            if np.all(p == 0) or np.any(np.isnan(p)):
                del all_psds[i]
                del all_times[i]
                if self._save_cov:
                    del all_cov[i]
                nrow -= 1
                nempty += 1
            else:
                i += 1

        if nempty > 0:
            print("Discarded {} empty or NaN psds".format(nempty), flush=True)

        # Throw away outlier PSDs by comparing the PSDs in specific bins

        if nrow < 10:
            nbad = 0
        else:
            all_good = np.isfinite(np.sum(all_psds, 1))
            for col in range(ncol - 1):
                if binfreq[col] < 0.001:
                    continue

                # Local outliers

                psdvalues = np.array([x[col] for x in all_psds])
                smooth_values = scipy.signal.medfilt(psdvalues, 11)
                good = np.ones(psdvalues.size, dtype=np.bool)
                good[psdvalues == 0] = False

                for i in range(10):
                    # Local test
                    diff = np.zeros(psdvalues.size)
                    diff[good] = np.log(psdvalues[good]) - np.log(smooth_values[good])
                    sdev = np.std(diff[good])
                    good[np.abs(diff) > 5 * sdev] = False
                    # Global test
                    diff = np.zeros(psdvalues.size)
                    diff[good] = np.log(psdvalues[good]) - np.mean(
                        np.log(psdvalues[good])
                    )
                    sdev = np.std(diff[good])
                    good[np.abs(diff) > 5 * sdev] = False

                all_good[np.logical_not(good)] = False

            bad = np.logical_not(all_good)
            nbad = np.sum(bad)
            if nbad > 0:
                for ii in np.argwhere(bad).ravel()[::-1]:
                    del all_psds[ii]
                    del all_times[ii]
                    if self._save_cov:
                        del all_cov[ii]

            if nbad > 0:
                print("Masked extra {} psds due to outliers." "".format(nbad))
        return all_psds, all_times, nempty + nbad, all_cov

    def save_psds(
        self, binfreq, all_psds, all_times, det1, det2, fsample, rootname, all_cov
    ):
        timer = Timer()
        timer.start()
        if det1 == det2:
            fn_out = os.path.join(self._out, "{}_{}.fits".format(rootname, det1))
        else:
            fn_out = os.path.join(
                self._out, "{}_{}_{}.fits".format(rootname, det1, det2)
            )
        all_psds = np.vstack([binfreq, all_psds])

        hdulist = [pf.PrimaryHDU()]

        cols = []
        cols.append(pf.Column(name="OBT", format="D", array=all_times))
        coldefs = pf.ColDefs(cols)
        hdu1 = pf.BinTableHDU.from_columns(coldefs)
        hdu1.header["RATE"] = fsample, "Sampling rate"
        hdulist.append(hdu1)

        cols = []
        cols.append(
            pf.Column(name="PSD", format="{}E".format(binfreq.size), array=all_psds)
        )
        coldefs = pf.ColDefs(cols)
        hdu2 = pf.BinTableHDU.from_columns(coldefs)
        hdu2.header["EXTNAME"] = str(det1), "Detector"
        hdu2.header["DET1"] = str(det1), "Detector1"
        hdu2.header["DET2"] = str(det2), "Detector2"
        hdulist.append(hdu2)

        if self._save_cov:
            all_cov = np.array(all_cov)
            cols = []
            nrow, ncol, nsamp = np.shape(all_cov)
            cols.append(
                pf.Column(
                    name="HITS",
                    format="{}J".format(nsamp),
                    array=np.ascontiguousarray(all_cov[:, 0, :]),
                )
            )
            cols.append(
                pf.Column(
                    name="COV",
                    format="{}E".format(nsamp),
                    array=np.ascontiguousarray(all_cov[:, 1, :]),
                )
            )
            coldefs = pf.ColDefs(cols)
            hdu3 = pf.BinTableHDU.from_columns(coldefs)
            hdu3.header["EXTNAME"] = str(det1), "Detector"
            hdu3.header["DET1"] = str(det1), "Detector1"
            hdu3.header["DET2"] = str(det2), "Detector2"
            hdulist.append(hdu3)

        hdulist = pf.HDUList(hdulist)

        with open(fn_out, "wb") as fits_out:
            hdulist.writeto(fits_out, overwrite=True)

        print("Detector {} vs. {} PSDs stored in {}".format(det1, det2, fn_out))
        return

    def process_downsampled_noise_estimate(
        self,
        timestamps,
        fsample,
        signal1,
        signal2,
        flags,
        gapflags_nsum,
        local_intervals,
        my_psds1,
        my_cov1,
        comm,
    ):
        # Get another PSD for a down-sampled TOD to measure the
        # low frequency power

        timestamps_decim = timestamps[:: self._nsum]
        # decimate() will smooth and downsample the signal in
        # each valid interval separately
        signal1_decim, flags_decim = self.decimate(
            signal1, flags, gapflags_nsum, local_intervals
        )
        if signal2 is not None:
            signal2_decim, flags_decim = self.decimate(
                signal2, flags, gapflags_nsum, local_intervals
            )

        if signal2 is None:
            result = autocov_psd(
                timestamps_decim,
                signal1_decim,
                flags_decim,
                min(self._lagmax, timestamps_decim.size),
                self._stationary_period,
                fsample / self._nsum,
                comm=comm,
                return_cov=self._save_cov,
            )
        else:
            result = crosscov_psd(
                timestamps_decim,
                signal1_decim,
                signal2_decim,
                flags_decim,
                min(self._lagmax, timestamps_decim.size),
                self._stationary_period,
                fsample / self._nsum,
                comm=comm,
                return_cov=self._save_cov,
            )
        if self._save_cov:
            my_psds2, my_cov2 = result
        else:
            my_psds2, my_cov2 = result, None

        # Ensure the two sets of PSDs are of equal length

        my_new_psds1 = []
        my_new_psds2 = []
        if self._save_cov:
            my_new_cov1 = []
            my_new_cov2 = []
        i = 0
        while i < min(len(my_psds1), len(my_psds2)):
            t1 = my_psds1[i][0]
            t2 = my_psds2[i][0]
            if np.isclose(t1, t2):
                my_new_psds1.append(my_psds1[i])
                my_new_psds2.append(my_psds2[i])
                if self._save_cov:
                    my_new_cov1.append(my_cov1[i])
                    my_new_cov2.append(my_cov2[i])
                i += 1
            else:
                if t1 < t2:
                    del my_psds1[i]
                    if self._cov:
                        del my_cov1[i]
                else:
                    del my_psds2[i]
                    if self._cov:
                        del my_cov1[i]
        my_psds1 = my_new_psds1
        my_psds2 = my_new_psds2
        if self._save_cov:
            my_cov1 = my_new_cov1
            my_cov2 = my_new_cov2

        if len(my_psds1) != len(my_psds2):
            while my_psds1[-1][0] > my_psds2[-1][0]:
                del my_psds1[-1]
                if self._save_cov:
                    del my_cov1[-1]
            while my_psds1[-1][0] < my_psds2[-1][0]:
                del my_psds2[-1]
                if self._save_cov:
                    del my_cov2[-1]
        return my_psds1, my_cov1, my_psds2, my_cov2

    def process_noise_estimate(
        self,
        signal1,
        signal2,
        flags,
        gapflags,
        gapflags_nsum,
        timestamps,
        fsample,
        comm,
        fileroot,
        det1,
        det2,
        local_intervals,
    ):
        # High pass filter the signal to avoid aliasing
        # self.highpass(signal1, noise_flags)
        # self.highpass(signal2, noise_flags)

        # Compute the autocovariance function and the matching
        # PSD for each stationary interval

        rank = 0
        if comm is not None:
            rank = comm.rank

        timer = Timer()
        timer.start()
        if signal2 is None:
            result = autocov_psd(
                timestamps,
                signal1,
                flags + gapflags,
                self._lagmax,
                self._stationary_period,
                fsample,
                comm=comm,
                return_cov=self._save_cov,
            )
        else:
            result = crosscov_psd(
                timestamps,
                signal1,
                signal2,
                flags + gapflags,
                self._lagmax,
                self._stationary_period,
                fsample,
                comm=comm,
                return_cov=self._save_cov,
            )
        if self._save_cov:
            my_psds1, my_cov1 = result
        else:
            my_psds1, my_cov1 = result, None

        if self._nsum > 1:
            (
                my_psds1,
                my_cov1,
                my_psds2,
                my_cov2,
            ) = self.process_downsampled_noise_estimate(
                timestamps,
                fsample,
                signal1,
                signal2,
                flags,
                gapflags_nsum,
                local_intervals,
                my_psds1,
                my_cov1,
                comm,
            )

        if rank == 0:
            timer.report_clear("Compute Correlators and PSDs")

        # Now bin the PSDs

        fmin = 1 / self._stationary_period
        fmax = fsample / 2

        my_binned_psds1, my_times1, binfreq10 = self.bin_psds(my_psds1, fmin, fmax)
        if self._nsum > 1:
            my_binned_psds2, _, binfreq20 = self.bin_psds(my_psds2, fmin, fmax)
        if rank == 0:
            timer.report_clear("Bin PSDs")

        # concatenate

        if binfreq10 is None:
            my_times = []
            my_binned_psds = []
            binfreq0 = None
        else:
            my_times = my_times1
            if self._save_cov:
                my_cov = my_cov1  # Only store the fully sampled covariance
            if self._nsum > 1:
                # frequencies that are usable in the down-sampled PSD
                fcut = fsample / 2 / self._naverage / 100
                ind1 = binfreq10 > fcut
                ind2 = binfreq20 <= fcut
                binfreq0 = np.hstack([binfreq20[ind2], binfreq10[ind1]])
                my_binned_psds = []
                for psd1, psd2 in zip(my_binned_psds1, my_binned_psds2):
                    my_binned_psds.append(np.hstack([psd2[ind2], psd1[ind1]]))
            else:
                binfreq0 = binfreq10
                my_binned_psds = my_binned_psds1

        # Collect and write the PSDs.  Start by determining the first
        # process to have a valid PSD to determine binning

        have_bins = binfreq0 is not None
        have_bins_all = None
        if comm is None:
            have_bins_all = [have_bins]
        else:
            have_bins_all = comm.allgather(have_bins)
        root = 0
        if np.any(have_bins_all):
            while not have_bins_all[root]:
                root += 1
        else:
            raise RuntimeError("None of the processes have valid PSDs")
        binfreq = None
        if comm is None:
            binfreq = binfreq0
        else:
            binfreq = comm.bcast(binfreq0, root=root)
        if binfreq0 is not None and np.any(binfreq != binfreq0):
            raise Exception(
                "{:4} : Binned PSD frequencies change. len(binfreq0)={}"
                ", len(binfreq)={}, binfreq0={}, binfreq={}. "
                "len(my_psds)={}".format(
                    rank, binfreq0.size, binfreq.size, binfreq0, binfreq, len(my_psds1)
                )
            )
        if len(my_times) != len(my_binned_psds):
            raise Exception(
                "ERROR: Process {} has len(my_times) = {}, len(my_binned_psds)"
                " = {}".format(rank, len(my_times), len(my_binned_psds))
            )
        all_times = None
        all_psds = None
        if comm is None:
            all_times = [my_times]
            all_psds = [my_binned_psds]
        else:
            all_times = comm.gather(my_times, root=0)
            all_psds = comm.gather(my_binned_psds, root=0)
        all_cov = None
        if self._save_cov:
            if comm is None:
                all_cov = [my_cov]
            else:
                all_cov = comm.gather(my_cov, root=0)
        if rank == 0:
            timer.report_clear("Collect PSDs")

        if rank == 0:
            # FIXME: original code had no timing report here for the previous block.
            # Was one intended?
            if len(all_times) != len(all_psds):
                raise Exception(
                    "ERROR: Process {} has len(all_times) = {}, len(all_psds)"
                    " = {} before deglitch".format(rank, len(all_times), len(all_psds))
                )
            # De-glitch the binned PSDs and write them to file
            i = 0
            while i < len(all_times):
                if len(all_times[i]) == 0:
                    del all_times[i]
                    del all_psds[i]
                    if self._save_cov:
                        del all_cov[i]
                else:
                    i += 1

            if len(all_times) != len(all_psds):
                raise Exception(
                    "ERROR: Process {} has len(all_times) = {}, len(all_psds)"
                    " = {} AFTER deglitch".format(rank, len(all_times), len(all_psds))
                )

            all_times = list(np.hstack(all_times))
            all_psds = list(np.hstack(all_psds))
            if self._save_cov:
                all_cov = list(np.hstack(all_cov))

            good_psds, good_times, nbad, good_cov = self.discard_outliers(
                binfreq, all_psds, all_times, all_cov
            )
            timer.report_clear("Discard outliers")

            self.save_psds(
                binfreq, all_psds, all_times, det1, det2, fsample, fileroot, all_cov
            )

            if nbad > 0:
                self.save_psds(
                    binfreq,
                    good_psds,
                    good_times,
                    det1,
                    det2,
                    fsample,
                    fileroot + "_good",
                    good_cov,
                )
            timer.report_clear("Write PSDs")

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np


class Noise(object):
    """Noise objects act as containers for noise PSDs.

    Noise is a base class for an object that describes the noise
    properties of all detectors for a single observation.

    Args:
        detectors (list): Names of detectors.
        freqs (dict): Dictionary of arrays of frequencies for `psds`.
        psds (dict): Dictionary of arrays which contain the PSD values
            for each detector or `mixmatrix` key.
        mixmatrix (dict): Mixing matrix describing how the PSDs should
            be combined for detector noise.  If provided, must contain
            entries for every detector, and every key specified for a
            detector must be defined in `freqs` and `psds`.
        indices (dict):  Integer index for every PSD, useful for
            generating indepedendent and repeateable noise realizations.
            If absent, running indices will be assigned and provided.

    Attributes:
        detectors (list): List of detector names
        keys (list): List of PSD names

    Raises:
        KeyError: If `freqs`, `psds`, `mixmatrix` or `indices` do not
            include all relevant entries.
        ValueError: If vector lengths in `freqs` and `psds` do not match.

    """

    def __init__(self, *, detectors, freqs, psds, mixmatrix=None, indices=None):

        self._dets = list(sorted(detectors))
        if mixmatrix is None:
            # Default diagonal mixing matrix
            self._keys = self._dets
            self._mixmatrix = None
        else:
            # Assemble the list of keys needed for the specified detectors
            keys = set()
            self._mixmatrix = {}
            for det in self._dets:
                self._mixmatrix[det] = {}
                for key, weight in mixmatrix[det].items():
                    keys.add(key)
                    self._mixmatrix[det][key] = weight
            self._keys = list(sorted(keys))
        if indices is None:
            self._indices = {}
            for i, key in enumerate(self._keys):
                self._indices[key] = i
        else:
            self._indices = dict(indices)
        self._freqs = {}
        self._psds = {}
        self._rates = {}

        for key in self._keys:
            if psds[key].shape[0] != freqs[key].shape[0]:
                raise ValueError("PSD length must match the number of frequencies")
            self._freqs[key] = np.copy(freqs[key])
            self._psds[key] = np.copy(psds[key])
            # last frequency point should be Nyquist
            self._rates[key] = 2.0 * self._freqs[key][-1]

    @property
    def detectors(self):
        """(list): list of strings containing the detector names."""
        return self._dets

    @property
    def keys(self):
        """(list): list of strings containing the PSD names."""
        return self._keys

    def multiply_ntt(self, key, data):
        """Filter the data with noise covariance."""
        raise NotImplementedError("multiply_ntt not yet implemented")

    def multiply_invntt(self, key, data):
        """Filter the data with inverse noise covariance."""
        raise NotImplementedError("multiply_invntt not yet implemented")

    def weight(self, det, key):
        """Return the mixing weight for noise `key` in `det`.

        Args:
            det (str): Detector name
            key (std): Mixing matrix key.
        Returns:
            weight (float): Mixing matrix weight

        """
        weight = 0
        if self._mixmatrix is None:
            if det == key:
                weight = 1
        elif key in self._mixmatrix[det]:
            weight = self._mixmatrix[det][key]
        return weight

    def index(self, key):
        """Return the PSD index for `key`

        Args:
            key (std): Detector name or mixing matrix key.
        Returns:
            index (int): PSD index.

        """
        return self._indices[key]

    def freq(self, key):
        """Get the frequencies corresponding to `key`.

        Args:
            key (str): Detector name or mixing matrix key.
        Returns:
            (array): Frequency bins that are used for the PSD.

        """
        return self._freqs[key]

    def rate(self, key):
        """Get the sample rate for `key`.

        Args:
            key (str): the detector name or mixing matrix key.
        Returns:
            (float): the sample rate in Hz.

        """
        return self._rates[key]

    def psd(self, key):
        """Get the PSD corresponding to `key`.

        Args:
            key (str): Detector name or mixing matrix key.
        Returns:
            (array): PSD matching the key.

        """
        return self._psds[key]



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import copy
import os
import re

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..todmap import OpMapMaker
from ..tod import OpPolyFilter


def add_mapmaker_args(parser):
    """Add mapmaker arguments"""
    parser.add_argument(
        "--mapmaker-prefix",
        required=False,
        default="toast",
        help="Output map prefix",
        dest="mapmaker_prefix",
    )
    parser.add_argument(
        "--mapmaker-mask",
        required=False,
        help="Destriping mask",
        dest="mapmaker_mask",
    )
    parser.add_argument(
        "--mapmaker-weightmap",
        required=False,
        help="Destriping weight map",
        dest="mapmaker_weightmap",
    )
    parser.add_argument(
        "--mapmaker-iter-max",
        required=False,
        default=1000,
        type=np.int,
        help="Maximum number of CG iterations",
        dest="mapmaker_iter_max",
    )
    parser.add_argument(
        "--mapmaker-precond-width",
        required=False,
        default=100,
        type=np.int,
        help="Width of the Madam band preconditioner",
        dest="mapmaker_precond_width",
    )
    parser.add_argument(
        "--mapmaker-baseline-length",
        required=False,
        default=10000.0,
        type=np.float,
        help="Destriping baseline length (seconds)",
        dest="mapmaker_baseline_length",
    )
    parser.add_argument(
        "--mapmaker-prefilter-order",
        required=False,
        type=np.int,
        help="Polynomial prefiltering for mapmaker",
        dest="mapmaker_prefilter_order",
    )
    parser.add_argument(
        "--mapmaker-fourier2D-order",
        required=False,
        type=np.int,
        help="Per sample 2D Fourier template order",
        dest="mapmaker_fourier2D_order",
    )
    parser.add_argument(
        "--mapmaker-fourier2D-subharmonics",
        required=False,
        action="store_true",
        help="Fit linear modes along with Fourier templates",
        dest="mapmaker_fourier2D_subharmonics",
    )
    parser.add_argument(
        "--no-mapmaker-fourier2D-subharmonics",
        required=False,
        action="store_false",
        help="Do not fit linear modes along with Fourier templates",
        dest="mapmaker_fourier2D_subharmonics",
    )
    parser.set_defaults(mapmaker_fourier2D_subharmonics=False)
    parser.add_argument(
        "--mapmaker-noisefilter",
        required=False,
        default=False,
        action="store_true",
        help="Destripe with the noise filter enabled",
        dest="mapmaker_noisefilter",
    )
    try:
        parser.add_argument(
            "--binmap",
            required=False,
            action="store_true",
            help="Write binned maps [default]",
            dest="write_binmap",
        )
        parser.add_argument(
            "--no-binmap",
            required=False,
            action="store_false",
            help="Do not write binned maps",
            dest="write_binmap",
        )
        parser.set_defaults(write_binmap=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--hits",
            required=False,
            action="store_true",
            help="Write hit maps [default]",
            dest="write_hits",
        )
        parser.add_argument(
            "--no-hits",
            required=False,
            action="store_false",
            help="Do not write hit maps",
            dest="write_hits",
        )
        parser.set_defaults(write_hits=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--wcov",
            required=False,
            action="store_true",
            help="Write white noise covariance [default]",
            dest="write_wcov",
        )
        parser.add_argument(
            "--no-wcov",
            required=False,
            action="store_false",
            help="Do not write white noise covariance",
            dest="write_wcov",
        )
        parser.set_defaults(write_wcov=True)
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--wcov-inv",
            required=False,
            action="store_true",
            help="Write inverse white noise covariance [default]",
            dest="write_wcov_inv",
        )
        parser.add_argument(
            "--no-wcov-inv",
            required=False,
            action="store_false",
            help="Do not write inverse white noise covariance",
            dest="write_wcov_inv",
        )
        parser.set_defaults(write_wcov_inv=True)
    except argparse.ArgumentError:
        pass

    # `nside` may already be added
    try:
        parser.add_argument(
            "--nside", required=False, default=512, type=np.int, help="Healpix NSIDE"
        )
    except argparse.ArgumentError:
        pass
    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass

    try:
        parser.add_argument(
            "--zip",
            required=False,
            action="store_true",
            help="Compress the map outputs",
            dest="zip_maps",
        )
        parser.add_argument(
            "--no-zip",
            required=False,
            action="store_false",
            help="Do not compress the map outputs",
            dest="zip_maps",
        )
        parser.set_defaults(zip_maps=True)
    except argparse.ArgumentError:
        pass

    return


@function_timer
def apply_mapmaker(
    args,
    comm,
    data,
    outpath,
    cache_name,
    time_comms=None,
    telescope_data=None,
    first_call=True,
    extra_prefix=None,
    verbose=True,
    bin_only=False,
):
    log = Logger.get()
    timer = Timer()

    if outpath is None:
        outpath = args.out

    file_root = args.mapmaker_prefix
    if extra_prefix is not None:
        if len(file_root) > 0 and not file_root.endswith("_"):
            file_root += "_"
        file_root += "{}".format(extra_prefix)

    if time_comms is None:
        time_comms = [("all", comm.comm_world)]

    if telescope_data is None:
        telescope_data = [("all", data)]

    if not bin_only and args.mapmaker_prefilter_order is not None:
        timer.start()
        if comm.world_rank == 0 and verbose:
            print(
                "Applying polynomial prefilter, order = {}".format(
                    args.mapmaker_prefilter_order
                ),
                flush=True,
            )
        polyfilter = OpPolyFilter(
            order=args.mapmaker_prefilter_order,
            name=cache_name,
            common_flag_mask=args.common_flag_mask,
        )
        polyfilter.exec(data)
        timer.stop()
        if comm.world_rank == 0 and verbose:
            timer.report("Polynomial prefilter")

    for time_name, time_comm in time_comms:
        for tele_name, tele_data in telescope_data:

            write_hits = args.write_hits and first_call
            write_wcov_inv = args.write_wcov_inv and first_call
            write_wcov = args.write_wcov and first_call
            if bin_only:
                baseline_length = None
                write_binned = True
                write_destriped = False
                fourier2D_order = None
            else:
                baseline_length = args.mapmaker_baseline_length
                write_binned = args.write_binmap
                write_destriped = True
                fourier2D_order = args.mapmaker_fourier2D_order

            if len(time_name.split("-")) == 3:
                # Special rules for daily maps
                if not args.do_daymaps:
                    continue
                if len(telescope_data) > 1 and tele_name == "all":
                    # Skip daily maps over multiple telescopes
                    continue
                # Do not destripe daily maps
                baseline_length = None
                write_binned = True
                write_destriped = False

            timer.clear()
            timer.start()

            if len(file_root) > 0 and not file_root.endswith("_"):
                file_root += "_"
            prefix = "{}telescope_{}_time_{}_".format(file_root, tele_name, time_name)

            mapmaker = OpMapMaker(
                nside=args.nside,
                nnz=3,
                name=cache_name,
                outdir=outpath,
                outprefix=prefix,
                write_hits=(args.write_hits and first_call),
                zip_maps=args.zip_maps,
                write_wcov_inv=(args.write_wcov_inv and first_call),
                write_wcov=(args.write_wcov and first_call),
                write_binned=write_binned,
                write_destriped=write_destriped,
                write_rcond=True,
                rcond_limit=1e-3,
                baseline_length=baseline_length,
                maskfile=args.mapmaker_mask,
                weightmapfile=args.mapmaker_weightmap,
                common_flag_mask=args.common_flag_mask,
                flag_mask=1,
                intervals="intervals",
                subharmonic_order=None,
                fourier2D_order=fourier2D_order,
                fourier2D_subharmonics=args.mapmaker_fourier2D_subharmonics,
                iter_min=3,
                iter_max=args.mapmaker_iter_max,
                use_noise_prior=args.mapmaker_noisefilter,
                precond_width=args.mapmaker_precond_width,
                pixels="pixels",
            )

            mapmaker.exec(tele_data, time_comm)

            # User needs to set TOAST_FUNCTIME to see timing results
            if "TOAST_FUNCTIME" in os.environ and os.environ["TOAST_FUNCTIME"]:
                mapmaker.report_timing()

    if comm.world_rank == 0 and verbose:
        timer.report_clear("  OpMapMaker")

    return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from .atm import (
    add_atmosphere_args,
    simulate_atmosphere,
    scale_atmosphere_by_frequency,
    update_atmospheric_noise_weights,
)
from .binning import add_binner_args, init_binner, apply_binner
from .classes import Telescope, Focalplane
from .debug import add_debug_args
from .dipole import add_dipole_args, simulate_dipole
from .dist import add_dist_args, get_comm, get_time_communicators
from .export import add_tidas_args, output_tidas, add_spt3g_args, output_spt3g
from .filters import (
    add_polyfilter_args,
    add_polyfilter2D_args,
    apply_polyfilter,
    apply_polyfilter2D,
    add_groundfilter_args,
    apply_groundfilter,
)
from .gain import add_gainscrambler_args, scramble_gains
from .mapmaker import add_mapmaker_args, apply_mapmaker
from .madam import add_madam_args, setup_madam, apply_madam
from .noise import add_noise_args, simulate_noise, get_analytic_noise
from .pointing import add_pointing_args, expand_pointing
from .sky_signal import (
    add_sky_map_args,
    add_pysm_args,
    scan_sky_signal,
    simulate_sky_signal,
    add_conviqt_args,
    apply_conviqt,
    apply_weighted_conviqt,
)
from .sss import add_sss_args, simulate_sss
from .todground import (
    add_todground_args,
    get_elevation_noise,
    get_breaks,
    load_schedule,
    load_weather,
    Site,
    CES,
    Schedule,
)
from .todsatellite import add_todsatellite_args


def add_mc_args(parser):
    """Add Monte Carlo arguments"""
    parser.add_argument(
        "--MC-start",
        required=False,
        default=0,
        type=np.int,
        help="First Monte Carlo noise realization",
    )
    parser.add_argument(
        "--MC-count",
        required=False,
        default=1,
        type=np.int,
        help="Number of Monte Carlo noise realizations",
    )
    return


@function_timer
def add_signal(args, comm, data, prefix_out, prefix_in, purge=False, verbose=True):
    """Add signal from cache prefix `prefix_in` to cache prefix
    `prefix_out`.  If `prefix_out` does not exit, it is created.

    """
    if prefix_in == prefix_out or prefix_in is None or prefix_out is None:
        return
    log = Logger.get()
    if comm.world_rank == 0 and verbose:
        log.info("Adding signal from {} to {}." "".format(prefix_in, prefix_out))
        if purge:
            log.info("Purging {} after adding".format(prefix_in))
    timer = Timer()
    timer.start()
    for obs in data.obs:
        tod = obs["tod"]
        for det in tod.local_dets:
            cachename_in = "{}_{}".format(prefix_in, det)
            cachename_out = "{}_{}".format(prefix_out, det)
            ref_in = tod.cache.reference(cachename_in)
            if tod.cache.exists(cachename_out):
                ref_out = tod.cache.reference(cachename_out)
                ref_out += ref_in
            else:
                ref_out = tod.cache.put(cachename_out, ref_in)
            del ref_in, ref_out
        # Purge only after all detectors are added, just in case
        # any one is an alias
        if purge:
            for det in tod.local_dets:
                cachename_in = "{}_{}".format(prefix_in, det)
                tod.cache.clear(cachename_in)
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Add signal")
    return


@function_timer
def copy_signal(args, comm, data, cache_prefix_in, cache_prefix_out, verbose=True):
    """Copy the signal in `cache_prefix_in` to `cache_prefix_out`."""
    if cache_prefix_in == cache_prefix_out:
        return
    log = Logger.get()
    timer = Timer()
    timer.start()
    if comm.world_rank == 0 and verbose:
        log.info(
            "Copying signal from {} to {}" "".format(cache_prefix_in, cache_prefix_out)
        )
    cachecopy = OpCacheCopy(cache_prefix_in, cache_prefix_out, force=True)
    cachecopy.exec(data)
    if comm.world_rank == 0 and verbose:
        timer.report_clear("Copy signal")
    return



#!/usr/bin/env python3

import os
import sys
import stat
import re
import shutil

import argparse

import toml


def create_argparse_file(props, file):
    """
    Take a dictionary from the TOML and make an argparse compatible
    parameter file.
    """
    with open(file, "w") as f:
        for k, v in props.items():
            if isinstance(v, bool):
                f.write("--{}\n".format(k))
            else:
                f.write("--{}\n".format(k))
                f.write("{}\n".format(v))
    return


def create_telescope_commands(
    tele, focalplane, schedule, jobdir, mpi_launch, mpi_procs, mpi_nodes
):
    """
    Return lines of shell code to generate a focalplane and schedule for a telescope.
    """
    mpistr = "{}".format(mpi_launch)
    if mpi_procs != "":
        mpistr = "{} {} 1".format(mpistr, mpi_procs)
    if mpi_nodes != "":
        mpistr = "{} {} 1".format(mpistr, mpi_nodes)
    detpix = focalplane["minpix"]
    fp_par_file = "focalplane_{}_{}.par".format(tele, detpix)
    fp_file = "focalplane_{}_{}.pkl".format(tele, detpix)
    fp_root = "focalplane_{}".format(tele)

    create_argparse_file(focalplane, os.path.join(jobdir, fp_par_file))

    outstr = '# Create telescope model for "{}"\n'.format(tele)
    outstr = '{}if [ ! -e "{}" ]; then\n'.format(outstr, fp_file)
    outstr = '{}    echo "Creating focalplane {} ..."\n'.format(outstr, fp_file)
    outstr = '{}    {} toast_fake_focalplane.py @{} --out "{}"\n'.format(
        outstr, mpistr, fp_par_file, fp_root
    )
    outstr = "{}fi\n".format(outstr)

    if len(schedule) > 0:
        schd_par_file = "schedule_{}.par".format(tele)
        schd_file = "schedule_{}.txt".format(tele)
        create_argparse_file(schedule, os.path.join(jobdir, schd_par_file))
        outstr = '{}if [ ! -e "{}" ]; then\n'.format(outstr, schd_file)
        outstr = '{}    echo "Creating schedule {} ..."\n'.format(outstr, schd_file)
        outstr = '{}    {} toast_ground_schedule.py @{} --out "{}"\n'.format(
            outstr, mpistr, schd_par_file, schd_file
        )
        outstr = "{}fi\n".format(outstr)

    outstr = "{}\n".format(outstr)

    return outstr


def create_pysm_commands(
    mapfile,
    nside,
    bandcenter_ghz,
    bandwidth_ghz,
    beam_arcmin,
    coord,
    mpi_launch,
    mpi_procs,
    mpi_nodes,
):
    """
    Return lines of shell code to generate the precomputed input sky map.
    """
    mpistr = "{}".format(mpi_launch)
    if mpi_procs != "":
        mpistr = "{} {} 1".format(mpistr, mpi_procs)
    if mpi_nodes != "":
        mpistr = "{} {} 1".format(mpistr, mpi_nodes)

    outstr = "# Create sky model\n"
    outstr = '{}if [ ! -e "{}" ]; then\n'.format(outstr, mapfile)
    outstr = '{}    echo "Creating sky model {} ..."\n'.format(outstr, mapfile)
    outstr = '{}    {} ./pysm_sky.py --output "{}" --nside {} --bandcenter_ghz {} --bandwidth_ghz {} --beam_arcmin {} --coord {}\n'.format(
        outstr,
        mpistr,
        mapfile,
        nside,
        bandcenter_ghz,
        bandwidth_ghz,
        beam_arcmin,
        coord,
    )
    outstr = "{}fi\n".format(outstr)

    outstr = "{}\n".format(outstr)

    return outstr


def ceil_div(a, b):
    """Handy, from: https://stackoverflow.com/questions/14822184/
    """
    return -(-a // b)


def compute_job_size(
    max_nodes,
    node_mem,
    node_max_procs,
    omp_min_threads,
    obs_count,
    obs_mem,
    obs_max_procs,
    large_groups=False,
):
    """
    Given constraints of the system and the job, compute the process grouping.

    By default, this function tries to make the group size as small as possible based
    on memory constraints and then sets the number of groups to be exactly the number
    of observations (each group has one observation).

    If large_groups is True, then instead the starting point is to make the groups
    as large as possible given the maximum allowed for the observations.  The observations are then distributed among these large groups.

    Args:


    Returns:
        (tuple): the number of nodes, procs per node, threads per proc, and group size.

    """
    print(
        "      system has {} nodes, each with {} GB memory and up to {} procs".format(
            max_nodes, node_mem, node_max_procs
        )
    )
    print(
        "      {} observations, each requiring {} GB memory and using up to {} procs".format(
            obs_count, obs_mem, obs_max_procs
        )
    )
    nodes = None
    node_procs = node_max_procs
    omp_threads = omp_min_threads
    group_size = None
    total_procs = None

    if large_groups:
        print("      Using largest possible group size.")
        group_mem = None
        if obs_max_procs > node_procs:
            # Maximizing the observation processes requires multiple nodes
            print(
                "      Max observation procs > node procs, using one or more nodes per group"
            )
            group_nodes = obs_max_procs // node_procs
            group_size = group_nodes * node_procs
            group_mem = group_nodes * node_mem
        else:
            # See if we can fit multiple maximal observations on one node.
            print(
                "      Max observation procs <= node procs, using up to one node per group"
            )
            group_size = node_procs
            group_mem = node_mem
            while (group_size % 2 == 0) and (group_size >= (2 * obs_max_procs)):
                group_size = group_size // 2
                group_mem /= 2.0
        print(
            "      Largest group size = {} procs, {} procs per node".format(
                group_size, node_procs
            )
        )

        # If this largest group size exceeds the node count, reduce it until it fits
        while ceil_div(group_size, node_procs) > max_nodes:
            print(
                "      Group size of {} procs exceeds max node count ({}).  Adjusting to:".format(
                    group_size, max_nodes
                )
            )
            group_size -= node_procs
            group_mem -= node_mem
            print("        group size = {}".format(group_size))

        # Does this maximum group size have enough memory for even one observation?
        if group_mem < obs_mem:
            print(
                "      Required memory for one observation ({}) is greater than max ({}), skipping job".format(
                    obs_mem, group_mem
                )
            )
            return (None, None, None, None)

        # Compute the number of observations per group that will fit into memory
        print("      Group has {} GB of memory".format(group_mem))
        obs_per_group = int(group_mem / obs_mem)
        print("      Max observations per group = {}".format(obs_per_group))

        # Compute the number of groups needed to process all observations
        num_groups = ceil_div(obs_count, obs_per_group)
        print("      {} total groups required".format(num_groups))

        # The required number of nodes
        total_procs = num_groups * group_size

    else:
        print("      Using smallest possible group size.")
        # Using the max number of processes per node, compute the number of processes
        # needed for one observation based on memory requirements.
        # This is the initial guess of the group size.
        if obs_mem > node_mem:
            # One observation requires multiple nodes.  Round up to the nearest whole
            # node.
            print(
                "      Max observation mem > node mem, using one or more nodes per group"
            )
            group_size = node_procs * (1 + int(obs_mem / node_mem))
        else:
            # One observation requires less than a node.  Try to fit multiple
            # observations on one node.
            print(
                "      Max observation mem <= node mem, using up to one node per group"
            )
            group_mem = node_mem
            group_size = node_procs
            while (group_size % 2 == 0) and (group_mem > (2 * obs_mem)):
                group_size = group_size // 2
                group_mem = group_mem / 2
        print(
            "      Minimum group size = {} procs, {} procs per node".format(
                group_size, node_procs
            )
        )

        if ceil_div(group_size, node_procs) > max_nodes:
            print(
                "      Required nodes for one group ({}) is greater than max ({}), skipping job".format(
                    ceil_div(group_size, node_procs), max_nodes
                )
            )
            return (None, None, None, None)

        # If the node count (at the max processes per node) to achieve the
        # required memory results in too many processes for the observation,
        # Reduce the process count and increase threading as needed while
        # maintaining the number of nodes for memory requirements.

        while group_size > obs_max_procs:
            print(
                "      Group size ({}) exceeds observation limit ({}).  Adjusting to:".format(
                    group_size, obs_max_procs
                )
            )
            node_procs = node_procs // 2
            omp_threads *= 2
            group_size = group_size // 2
            print(
                "        {} procs per node ({} threads per proc), group size = {}".format(
                    node_procs, omp_threads, group_size
                )
            )

        # We now have the size of one group, assuming that each group has one
        # observation.  Compute the number of nodes needed.
        total_procs = obs_count * group_size

    nodes = ceil_div(total_procs, node_procs)
    if nodes > max_nodes:
        print(
            "      Required nodes ({}) is greater than max ({}), skipping job".format(
                nodes, max_nodes
            )
        )
        return (None, None, None, None)

    while node_procs > total_procs:
        # Our job is smaller than one node, reduce the size
        node_procs = node_procs // 2
        omp_threads *= 2

    print(
        "      Using {} nodes with {} procs per node, {} threads per proc, and {} procs per group".format(
            nodes, node_procs, omp_threads, group_size
        )
    )

    return (nodes, node_procs, omp_threads, group_size)


def main():
    parser = argparse.ArgumentParser(description="Generate TOAST example scripts.")

    parser.add_argument(
        "--config",
        required=False,
        default="config.toml",
        help="The input config file in TOML format",
    )

    parser.add_argument(
        "--datadir",
        required=False,
        default=None,
        help="The directory where you have fetched the data",
    )

    parser.add_argument(
        "--large_groups",
        required=False,
        default=False,
        action="store_true",
        help="Use largest groups possible, rather than smallest.",
    )

    # Location of this script
    example_dir = os.path.dirname(os.path.realpath(__file__))

    args = parser.parse_args()

    datadir = args.datadir
    if datadir is None:
        datadir = os.path.join(example_dir, "data")

    config = toml.load(args.config)

    systems = list(config["systems"].keys())

    jobtypes = list(config["jobs"].keys())

    for system in systems:
        print("Generating scripts for {}:".format(system))
        sysprops = config["systems"][system]
        template = sysprops["template"]
        _, template_ext = os.path.splitext(template)

        for job in jobtypes:
            job_props = config["jobs"][job]
            print("  Job type {}:".format(job))
            sizes = [x for x in job_props.keys() if x != "common"]

            # Parse common parameters
            common = job_props["common"]
            common_tele = common["telescopes"]
            common_pipeline = common["pipeline"]

            for sz in sizes:
                # The properties for this job size
                size_props = job_props[sz]

                print("    size {}:".format(sz))
                jobname = "{}_{}".format(job, sz)
                outdir = "job_{}_{}_{}".format(system, job, sz)
                os.makedirs(outdir, exist_ok=True)

                # link to data directory
                jobdata = os.path.join(outdir, "data")
                if os.path.islink(jobdata):
                    # a previously created link...
                    os.remove(jobdata)
                os.symlink(datadir, jobdata)

                # link scripts
                # src_py = os.path.join(example_dir, "trigger_astropy.py")
                # dest_py = os.path.join(outdir, "trigger_astropy.py")
                # if os.path.islink(dest_py):
                #     os.remove(dest_py)
                # os.symlink(src_py, dest_py)

                src_py = os.path.join(example_dir, "pysm_sky.py")
                dest_py = os.path.join(outdir, "pysm_sky.py")
                if os.path.islink(dest_py):
                    os.remove(dest_py)
                os.symlink(src_py, dest_py)

                telescopes = dict()
                focalplanes = list()
                schedules = list()
                for ctele, stele in zip(common_tele, size_props["telescopes"]):
                    tname = ctele["name"]
                    if tname != stele["name"]:
                        raise RuntimeError(
                            "common and per-size telescopes must have same names and order"
                        )
                    fp = dict(ctele["focalplane"])
                    if "focalplane" in stele:
                        fp.update(stele["focalplane"])
                    schd = dict()
                    if "schedule" in ctele:
                        schd.update(ctele["schedule"])
                    if "schedule" in stele:
                        schd.update(stele["schedule"])
                    telescopes[tname] = dict()
                    telescopes[tname]["focalplane"] = fp
                    telescopes[tname]["schedule"] = schd
                    focalplanes.append(
                        "focalplane_{}_{}.pkl".format(tname, fp["minpix"])
                    )
                    if len(schd) > 0:
                        schedules.append("schedule_{}.txt".format(tname))

                # Write pipeline parameter file
                pipeline = dict(common_pipeline)
                if "pipeline" in size_props:
                    pipeline.update(size_props["pipeline"])
                pipeline_file = os.path.join(outdir, "pipeline.par")
                create_argparse_file(pipeline, pipeline_file)

                # Compute the job size
                req = size_props["requirements"]

                (nodes, node_procs, omp_threads, group_size) = compute_job_size(
                    sysprops["max_nodes"],
                    sysprops["node_mem_gb"],
                    sysprops["node_max_procs"],
                    sysprops["omp_min_threads"],
                    req["obs_count"],
                    req["obs_mem_gb"],
                    req["obs_max_procs"],
                    large_groups=args.large_groups,
                )
                if nodes is None:
                    continue

                # Build up the job substitution dictionary.  Substitute our calculated
                # node parameters.

                jobsub = dict(sysprops)
                jobsub.update(req)
                jobsub["node_procs"] = node_procs
                jobsub["omp_threads"] = omp_threads
                jobsub["group_size"] = group_size
                jobsub["nodes"] = nodes
                jobsub["jobname"] = jobname
                jobsub["focalplane_list"] = "--focalplane {}".format(
                    ",".join(focalplanes)
                )
                jobsub["schedule_list"] = ""
                if len(schedules) > 0:
                    jobsub["schedule_list"] = "--schedule {}".format(
                        ",".join(schedules)
                    )

                pysm_fwhm = None
                pysm_bandcenter = None
                pysm_bandwidth = None
                for tele, tprops in telescopes.items():
                    fp = tprops["focalplane"]
                    if pysm_fwhm is None:
                        pysm_fwhm = fp["fwhm"]
                        pysm_bandcenter = fp["bandcenter_ghz"]
                        pysm_bandwidth = fp["bandwidth_ghz"]
                    else:
                        if pysm_fwhm != fp["fwhm"]:
                            raise RuntimeError(
                                "These jobs only support a single det type"
                            )
                        if pysm_bandcenter != fp["bandcenter_ghz"]:
                            raise RuntimeError(
                                "These jobs only support a single det type"
                            )
                        if pysm_bandwidth != fp["bandwidth_ghz"]:
                            raise RuntimeError(
                                "These jobs only support a single det type"
                            )

                # Read the template and write the job script
                outfile = os.path.join(outdir, "run{}".format(template_ext))
                with open(outfile, "w") as fout:
                    with open(template, "r") as fin:
                        for line in fin:
                            if re.match("@TELESCOPES@.*", line) is not None:
                                # Write our commands to create telescopes
                                line = ""
                                for tele, tprops in telescopes.items():
                                    tcom = create_telescope_commands(
                                        tele,
                                        tprops["focalplane"],
                                        tprops["schedule"],
                                        outdir,
                                        sysprops["mpi_launch"],
                                        sysprops["mpi_procs"],
                                        sysprops["mpi_nodes"],
                                    )
                                    line = "{}{}".format(line, tcom)
                            elif re.match("@PYSM_SKY@.*", line) is not None:
                                line = create_pysm_commands(
                                    pipeline["input-map"],
                                    pipeline["nside"],
                                    pysm_bandcenter,
                                    pysm_bandwidth,
                                    pysm_fwhm,
                                    pipeline["coord"],
                                    sysprops["mpi_launch"],
                                    sysprops["mpi_procs"],
                                    sysprops["mpi_nodes"],
                                )
                            else:
                                for k, v in jobsub.items():
                                    line = re.sub(
                                        "@{}@".format(k), "{}".format(v), line
                                    )
                            fout.write(line)
                if template_ext == ".sh":
                    st = os.stat(outfile)
                    os.chmod(outfile, st.st_mode | stat.S_IEXEC)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""Invert a block diagonal covariance matrix.
"""
import os
import re
import sys
import argparse
import traceback

import numpy as np

import healpy as hp

from toast.mpi import get_world

from toast.utils import Logger

from toast.dist import distribute_uniform

from toast.map import DistPixels, covariance_rcond


def main():
    log = Logger.get()

    parser = argparse.ArgumentParser(
        description="Read a toast covariance matrix and write the inverse condition number map"
    )

    parser.add_argument(
        "--input", required=True, default=None, help="The input covariance FITS file"
    )

    parser.add_argument(
        "--output",
        required=False,
        default=None,
        help="The output inverse condition map FITS file.",
    )

    try:
        args = parser.parse_args()
    except SystemExit:
        return

    # get options

    infile = args.input
    outfile = None
    if args.output is not None:
        outfile = args.output
    else:
        inmat = re.match(r"(.*)\.fits", infile)
        if inmat is None:
            log.error("input file should have .fits extension")
            return
        inroot = inmat.group(1)
        outfile = "{}_rcond.fits".format(inroot)

    # Get the default communicator
    mpiworld, procs, rank = get_world()

    # We need to read the header to get the size of the matrix.
    # This would be a trivial function call in astropy.fits or
    # fitsio, but we don't want to bring in a whole new dependency
    # just for that.  Instead, we open the file with healpy in memmap
    # mode so that nothing is actually read except the header.

    nside = 0
    nnz = 0
    if rank == 0:
        fake, head = hp.read_map(infile, h=True, memmap=True)
        for key, val in head:
            if key == "NSIDE":
                nside = int(val)
            if key == "TFIELDS":
                nnz = int(val)
    if mpiworld is not None:
        nside = mpiworld.bcast(nside, root=0)
        nnz = mpiworld.bcast(nnz, root=0)

    npix = 12 * nside ** 2
    subnside = int(nside / 16)
    if subnside == 0:
        subnside = 1
    subnpix = 12 * subnside ** 2
    nsubmap = int(npix / subnpix)

    # divide the submaps as evenly as possible among processes

    dist = distribute_uniform(nsubmap, procs)
    local = np.arange(dist[rank][0], dist[rank][0] + dist[rank][1])

    if rank == 0:
        if os.path.isfile(outfile):
            os.remove(outfile)

    if mpiworld is not None:
        mpiworld.barrier()

    # create the covariance and inverse condition number map

    cov = DistPixels(
        comm=mpiworld, dtype=np.float64, size=npix, nnz=nnz, submap=subnpix, local=local
    )

    # read the covariance
    log.info("Reading covariance {}".format(infile))
    cov.read_healpix_fits(infile)

    # every process computes its local piece
    log.info("Computing condition number map")
    rcond = covariance_rcond(cov)

    # write the map
    log.info("Writing condition number map")
    rcond.write_healpix_fits(outfile)
    return


if __name__ == "__main__":
    try:
        main()
    except:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from ..timing import function_timer

from ..fft import FFTPlanReal1DStore

from .tod_math import sim_noise_timestream

from ..op import Operator


class OpSimNoise(Operator):
    """Operator which generates noise timestreams.

    This passes through each observation and every process generates data
    for its assigned samples.  The dictionary for each observation should
    include a unique 'ID' used in the random number generation.  The
    observation dictionary can optionally include a 'global_offset' member
    that might be useful if you are splitting observations and want to
    enforce reproducibility of a given sample, even when using
    different-sized observations.

    Args:
        out (str): accumulate data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        realization (int): if simulating multiple realizations, the realization
            index.
        component (int): the component index to use for this noise simulation.
        noise (str): PSD key in the observation dictionary.

    """

    def __init__(
        self, out="noise", realization=0, component=0, noise="noise", rate=None
    ):
        # Call the parent class constructor.
        super().__init__()

        self._out = out
        self._oversample = 2
        self._realization = realization
        self._component = component
        self._noisekey = noise
        self._rate = rate

    @function_timer
    def exec(self, data):
        """Generate noise timestreams.

        This iterates over all observations and detectors and generates
        the noise timestreams based on the noise object for the current
        observation.

        Args:
            data (toast.Data): The distributed data.

        Raises:
            KeyError: If an observation in data does not have noise
                object defined under given key.
            RuntimeError: If observations are not split into chunks.

        """
        for obs in data.obs:
            obsindx = 0
            if "id" in obs:
                obsindx = obs["id"]
            else:
                print("Warning: observation ID is not set, using zero!")

            telescope = 0
            if "telescope" in obs:
                telescope = obs["telescope_id"]

            global_offset = 0
            if "global_offset" in obs:
                global_offset = obs["global_offset"]

            tod = obs["tod"]
            if self._noisekey in obs:
                nse = obs[self._noisekey]
            else:
                raise KeyError(
                    "Observation does not contain noise under "
                    '"{}"'.format(self._noisekey)
                )
            if tod.local_chunks is None:
                raise RuntimeError(
                    "noise simulation for uniform distributed "
                    "samples not implemented"
                )

            # eventually we'll redistribute, to allow long correlations...

            if self._rate is None:
                times = tod.local_times()
            else:
                times = None

            # Iterate over each chunk.

            chunk_first = tod.local_samples[0]
            for curchunk in range(tod.local_chunks[1]):
                chunk_first += self.simulate_chunk(
                    tod=tod,
                    nse=nse,
                    curchunk=curchunk,
                    chunk_first=chunk_first,
                    obsindx=obsindx,
                    times=times,
                    telescope=telescope,
                    global_offset=global_offset,
                )

        return

    @function_timer
    def simulate_chunk(
        self,
        *,
        tod,
        nse,
        curchunk,
        chunk_first,
        obsindx,
        times,
        telescope,
        global_offset
    ):
        """Simulate one chunk of noise for all detectors.

        Args:
            tod (toast.tod.TOD): TOD object for the observation.
            nse (toast.tod.Noise): Noise object for the observation.
            curchunk (int): The local index of the chunk to simulate.
            chunk_first (int): First global sample index of the chunk.
            obsindx (int): Observation index for random number stream.
            times (int): Timestamps for effective sample rate.
            telescope (int): Telescope index for random number stream.
            global_offset (int): Global offset for random number stream.

        Returns:
            chunk_samp (int): Number of simulated samples

        """
        chunk_samp = tod.total_chunks[tod.local_chunks[0] + curchunk]
        local_offset = chunk_first - tod.local_samples[0]

        if self._rate is None:
            # compute effective sample rate
            rate = 1 / np.median(
                np.diff(times[local_offset : local_offset + chunk_samp])
            )
        else:
            rate = self._rate

        for key in nse.keys:
            # Check if noise matching this PSD key is needed
            weight = 0.0
            for det in tod.local_dets:
                weight += np.abs(nse.weight(det, key))
            if weight == 0:
                continue

            # Simulate the noise matching this key
            nsedata = sim_noise_timestream(
                self._realization,
                telescope,
                self._component,
                obsindx,
                nse.index(key),
                rate,
                chunk_first + global_offset,
                chunk_samp,
                self._oversample,
                nse.freq(key),
                nse.psd(key),
            )

            # Add the noise to all detectors that have nonzero weights
            for det in tod.local_dets:
                weight = nse.weight(det, key)
                if weight == 0:
                    continue
                cachename = "{}_{}".format(self._out, det)
                if tod.cache.exists(cachename):
                    ref = tod.cache.reference(cachename)
                else:
                    ref = tod.cache.create(
                        cachename, np.float64, (tod.local_samples[1],)
                    )
                ref[local_offset : local_offset + chunk_samp] += weight * nsedata
                del ref

        # Release the work space allocated in the FFT plan store.
        store = FFTPlanReal1DStore.get()
        store.clear()

        return chunk_samp



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import warnings

import numpy as np

from scipy.constants import arcmin

import healpy as hp

from ..mpi import MPIShared

from ..timing import function_timer

from .._libtoast import fast_scanning_float32

DTYPE = np.float32


@function_timer
def plug_holes(m, verbose=False, in_place=True, nest=False):
    """Use simple downgrading to derive estimates of the missing pixel values"""
    nbad_start = np.sum(np.isclose(m, hp.UNSEEN))

    if nbad_start == m.size:
        if verbose:
            print("plug_holes: All map pixels are empty. Cannot plug holes", flush=True)
        return

    if nbad_start == 0:
        return

    nside = hp.get_nside(m)
    npix = m.size
    if nest:
        mnest = m.copy()
    else:
        mnest = hp.reorder(m, r2n=True)

    lowres = mnest
    nside_lowres = nside
    bad = np.isclose(mnest, hp.UNSEEN)
    while np.any(bad) and nside_lowres > 1:
        nside_lowres //= 2
        lowres = hp.ud_grade(lowres, nside_lowres, order_in="NESTED")
        hires = hp.ud_grade(lowres, nside, order_in="NESTED")
        bad = np.isclose(mnest, hp.UNSEEN)
        mnest[bad] = hires[bad]

    nbad_end = np.sum(bad)

    if nbad_end != 0:
        mn = np.mean(mnest[np.logical_not(bad)])
        mnest[bad] = mn

    if not in_place:
        m = m.copy()
    if nest:
        m[:] = mnest
    else:
        m[:] = hp.reorder(mnest, n2r=True)

    if verbose and nbad_start != 0:
        print(
            "plug_holes: Filled {} missing pixels ({:.2f}%), lowest "
            "resolution was Nside={}.".format(
                nbad_start, (100.0 * nbad_start) // npix, nside_lowres
            )
        )
    return m


class MapSampler:
    """
    MapSampler objects store maps in the node shared memory and allow
    bilinear interpolation of the maps into TOD.
    """

    @function_timer
    def __init__(
        self,
        map_path,
        pol=False,
        pol_fwhm=None,
        no_temperature=False,
        dtype=None,
        verbose=False,
        nside=None,
        comm=None,
        cache=None,
        preloaded_map=None,
        buflen=1000000,
        nest=False,
    ):
        """
        Instantiate the map sampler object, load a healpix
        map in a file located at map_path

        if pol==True, reads I,Q,U maps from extensions 0, 1, 2
        """

        if not pol and no_temperature:
            raise RuntimeError("You cannot have pol=False, " "no_temperature=True")

        self.path = map_path
        self.pol = pol
        self.pol_fwhm = pol_fwhm
        self._map = None
        self._map_Q = None
        self._map_U = None
        self.nest = nest
        if nest:
            self.order = "NESTED"
        else:
            self.order = "RING"
        self.buflen = buflen
        # Output data type, internal is always DTYPE
        if dtype is not None:
            warnings.warn("MapSampler no longer supports dtype", DeprecationWarning)

        # Use healpy to load the map into memory.

        if comm is None:
            self.comm = None
            self.rank = 0
            self.ntask = 1
        else:
            self.comm = comm
            self.rank = comm.Get_rank()
            self.ntask = comm.Get_size()

        self.shmem = self.ntask > 1
        self.pol = pol

        if self.rank == 0:
            if self.pol:
                if preloaded_map is not None:
                    if no_temperature:
                        (self._map_Q, self._map_U) = np.array(
                            preloaded_map, dtype=DTYPE
                        )
                    else:
                        (self._map, self._map_Q, self._map_U) = np.array(
                            preloaded_map, dtype=DTYPE
                        )
                else:
                    if no_temperature:
                        self._map_Q, self._map_U = hp.read_map(
                            self.path,
                            field=[1, 2],
                            dtype=DTYPE,
                            verbose=verbose,
                            memmmap=True,
                            nest=self.nest,
                        )
                    else:
                        try:
                            self._map, self._map_Q, self._map_U = hp.read_map(
                                self.path,
                                field=[0, 1, 2],
                                dtype=DTYPE,
                                verbose=verbose,
                                memmap=True,
                                nest=self.nest,
                            )
                        except IndexError:
                            print(
                                "WARNING: {} is not polarized".format(self.path),
                                flush=True,
                            )
                            self.pol = False
                            self._map = hp.read_map(
                                self.path,
                                dtype=DTYPE,
                                verbose=verbose,
                                memmap=True,
                                nest=self.nest,
                            )

                if nside is not None:
                    if not no_temperature:
                        self._map = hp.ud_grade(
                            self._map,
                            nside,
                            dtype=DTYPE,
                            order_in=self.order,
                            order_out=self.order,
                        )
                    if self.pol:
                        self._map_Q = hp.ud_grade(
                            self._map_Q,
                            nside,
                            dtype=DTYPE,
                            order_in=self.order,
                            order_out=self.order,
                        )
                        self._map_U = hp.ud_grade(
                            self._map_U,
                            nside,
                            dtype=DTYPE,
                            order_in=self.order,
                            order_out=self.order,
                        )

                if self.pol_fwhm is not None:
                    if not no_temperature:
                        plug_holes(self._map, verbose=verbose, nest=self.nest)
                    if self.pol:
                        plug_holes(self._map_Q, verbose=verbose, nest=self.nest)
                        plug_holes(self._map_U, verbose=verbose, nest=self.nest)
            else:
                if preloaded_map is not None:
                    self._map = np.array(preloaded_map, dtype=DTYPE)
                else:
                    self._map = hp.read_map(
                        map_path,
                        field=[0],
                        dtype=DTYPE,
                        verbose=verbose,
                        memmap=True,
                        nest=self.nest,
                    )
                if nside is not None:
                    self._map = hp.ud_grade(
                        self._map,
                        nside,
                        dtype=DTYPE,
                        order_in=self.order,
                        order_out=self.order,
                    )
                plug_holes(self._map, verbose=verbose, nest=self.nest)

        if self.ntask > 1:
            self.pol = comm.bcast(self.pol, root=0)
            npix = 0
            if self.rank == 0:
                if self.pol:
                    npix = len(self._map_Q)
                else:
                    npix = len(self._map)
            npix = comm.bcast(npix, root=0)
            if self.shmem:
                shared = MPIShared((npix,), np.dtype(DTYPE), comm)
                if not no_temperature:
                    shared.set(self._map, (0,), fromrank=0)
                    self._map = shared
                if self.pol:
                    shared_Q = MPIShared((npix,), np.dtype(DTYPE), comm)
                    shared_Q.set(self._map_Q, (0,), fromrank=0)
                    self._map_Q = shared_Q
                    shared_U = MPIShared((npix,), np.dtype(DTYPE), comm)
                    shared_U.set(self._map_U, (0,), fromrank=0)
                    self._map_U = shared_U
            else:
                if self.rank != 0:
                    if not no_temperature:
                        self._map = np.zeros(npix, dtype=DTYPE)
                    if self.pol:
                        self._map_Q = np.zeros(npix, dtype=DTYPE)
                        self._map_U = np.zeros(npix, dtype=DTYPE)

                if not no_temperature:
                    comm.Bcast(self._map, root=0)
                if self.pol:
                    comm.Bcast(self._map_Q, root=0)
                    comm.Bcast(self._map_U, root=0)

        if self.pol:
            self.npix = len(self._map_Q[:])
        else:
            self.npix = len(self._map[:])
        self.nside = hp.npix2nside(self.npix)

        self.cache = cache
        self.instance = 0
        if self.cache is not None and not self.shmem:
            # Increase the instance counter until we find an unused
            # instance.  If the user did not want to store duplicates,
            # they would not have created two identical mapsampler
            # objects.
            while self.cache.exists(self._cachename("I")):
                self.instance += 1
            if not no_temperature:
                self._map = self.cache.put(self._cachename("I"), self._map)
            if self.pol:
                self._map_Q = self.cache.put(self._cachename("Q"), self._map_Q)
                self._map_U = self.cache.put(self._cachename("U"), self._map_U)

        if self.pol_fwhm is not None:
            self.smooth(self.pol_fwhm, pol_only=True)
        return

    @function_timer
    def smooth(self, fwhm, lmax=None, pol_only=False):
        """Smooth the map with a Gaussian kernel."""
        if self.rank == 0:
            if pol_only:
                print(
                    "Smoothing the polarization to {} arcmin".format(fwhm), flush=True
                )
            else:
                print("Smoothing the map to {} arcmin".format(fwhm), flush=True)

        if lmax is None:
            lmax = min(np.int(fwhm / 60 * 512), 2 * self.nside)

        # If the map is in node-shared memory, only the root process on each
        # node does the smoothing.
        if not self.shmem or self._map.nodecomm.rank == 0:
            if self.pol:
                m = np.vstack([self._map[:], self._map_Q[:], self._map_U[:]])
            else:
                m = self._map[:]
            if self.nest:
                m = hp.reorder(m, n2r=True)
            smap = hp.smoothing(m, fwhm=fwhm * arcmin, lmax=lmax, verbose=False)
            del m
            if self.nest:
                smap = hp.reorder(smap, r2n=True)
        else:
            # Convenience dummy variable
            smap = np.zeros([3, 12])

        if not pol_only:
            if self.shmem:
                self._map.set(smap[0].astype(DTYPE), (0,), fromrank=0)
            else:
                self._map[:] = smap[0]

        if self.pol:
            if self.shmem:
                self._map_Q.set(smap[1].astype(DTYPE), (0,), fromrank=0)
                self._map_U.set(smap[2].astype(DTYPE), (0,), fromrank=0)
            else:
                self._map_Q[:] = smap[1]
                self._map_U[:] = smap[2]

        self.pol_fwhm = fwhm
        return

    def _cachename(self, stokes):
        """
        Construct a cache name string for the selected Stokes map
        """
        return "{}_ns{:04}_{}_{:04}".format(
            self.path, self.nside, stokes, self.instance
        )

    @function_timer
    def __del__(self):
        """
        Explicitly free memory taken up in the cache.
        """
        if self.cache is not None:
            if self._map is not None:
                del self._map
                if not self.shmem:
                    self.cache.destroy(self._cachename("I"))
            if self.pol:
                del self._map_Q
                del self._map_U
                if not self.shmem:
                    self.cache.destroy(self._cachename("Q"))
                    self.cache.destroy(self._cachename("U"))
        return

    @function_timer
    def __iadd__(self, other):
        """Accumulate provided Mapsampler object with this one."""
        if self.shmem:
            # One process does the manipulation on each node
            self._map._nodecomm.Barrier()
            if self._map._noderank == 0:
                self._map._data[:] += other._map[:]
            if self.pol and other.pol:
                if self._map_Q._noderank == (1 % self._map_Q._nodeprocs):
                    self._map_Q._data[:] += other._map_Q[:]
                if self._map_U._noderank == (2 % self._map_U._nodeprocs):
                    self._map_U._data[:] += other._map_U[:]
            self._map._nodecomm.Barrier()
        else:
            self._map += other._map
            if self.pol and other.pol:
                self._map_Q += other._map_Q
                self._map_U += other._map_U
        return self

    @function_timer
    def __isub__(self, other):
        """Subtract provided Mapsampler object from this one."""
        if self.shmem:
            # One process does the manipulation on each node
            self._map._nodecomm.Barrier()
            if self._map._noderank == 0:
                self._map._data[:] -= other._map[:]
            if self.pol and other.pol:
                if self._map_Q._noderank == (1 % self._map_Q._nodeprocs):
                    self._map_Q._data[:] -= other._map_Q[:]
                if self._map_U._noderank == (2 % self._map_U._nodeprocs):
                    self._map_U._data[:] -= other._map_U[:]
            self._map._nodecomm.Barrier()
        else:
            self._map -= other._map
            if self.pol and other.pol:
                self._map_Q -= other._map_Q
                self._map_U -= other._map_U
        return self

    @function_timer
    def __imul__(self, other):
        """Scale the maps in this MapSampler object"""
        if self.shmem:
            # One process does the manipulation on each node
            self._map._nodecomm.Barrier()
            if self._map._noderank == 0:
                self._map._data[:] *= other
            if self.pol:
                if self._map_Q._noderank == (1 % self._map_Q._nodeprocs):
                    self._map_Q._data[:] *= other
                if self._map_U._noderank == (2 % self._map_U._nodeprocs):
                    self._map_U._data[:] *= other
            self._map._nodecomm.Barrier()
        else:
            self._map *= other
            if self.pol:
                self._map_Q *= other
                self._map_U *= other
        return self

    @function_timer
    def __itruediv__(self, other):
        """Divide the maps in this MapSampler object"""
        if self.shmem:
            self._map._nodecomm.Barrier()
            if self._map._noderank == 0:
                self._map._data[:] /= other
            if self.pol:
                if self._map_Q._noderank == (1 % self._map_Q._nodeprocs):
                    self._map_Q._data[:] /= other
                if self._map_U._noderank == (2 % self._map_U._nodeprocs):
                    self._map_U._data[:] /= other
            self._map._nodecomm.Barrier()
        else:
            self._map /= other
            if self.pol:
                self._map_Q /= other
                self._map_U /= other
        return self

    @function_timer
    def at(self, theta, phi, interp_pix=None, interp_weights=None):
        """
        Use healpy bilinear interpolation to interpolate the
        map.  User must make sure that coordinate system used
        for theta and phi matches the map coordinate system.
        """
        if self._map is None:
            raise RuntimeError("No temperature map to sample")

        n = len(theta)
        stepsize = self.buflen
        signal = np.zeros(n, dtype=np.float32)

        # DEBUG begin
        if np.any(theta < 0) or np.any(theta > np.pi):
            raise RuntimeError("bad theta")
        if np.any(phi < 0) or np.any(phi > 2 * np.pi):
            raise RuntimeError("bad phi")
        # DEBUG end

        for istart in range(0, n, stepsize):
            istop = min(istart + stepsize, n)
            ind = slice(istart, istop)
            if interp_pix is None or interp_weights is None:
                p, w = hp.get_interp_weights(
                    self.nside, theta[ind], phi[ind], nest=self.nest
                )
            else:
                p = np.ascontiguousarray(interp_pix[:, ind])
                w = np.ascontiguousarray(interp_weights[:, ind])
            buf = np.zeros(istop - istart, dtype=np.float64)
            fast_scanning_float32(buf, p, w, self._map[:])
            signal[ind] = buf
        return signal

    @function_timer
    def atpol(
        self,
        theta,
        phi,
        IQUweight,
        onlypol=False,
        interp_pix=None,
        interp_weights=None,
        pol=True,
        pol_deriv=False,
    ):
        """
        Use healpy bilinear interpolation to interpolate the
        map.  User must make sure that coordinate system used
        for theta and phi matches the map coordinate system.
        IQUweight is an array of shape (nsamp,3) returned by the
        pointing library that gives the weights of the I,Q, and U maps.

        Args:
            pol_deriv(bool):  Return the polarization angle derivative
                of the signal instead of the actual signal.

        """
        if onlypol and not self.pol:
            return None

        if not self.pol or not pol:
            return self.at(
                theta, phi, interp_pix=interp_pix, interp_weights=interp_weights
            )

        if np.shape(IQUweight)[1] != 3:
            raise RuntimeError(
                "Cannot sample polarized map with only " "intensity weights"
            )

        n = len(theta)
        stepsize = self.buflen
        signal = np.zeros(n, dtype=np.float32)

        for istart in range(0, n, stepsize):
            istop = min(istart + stepsize, n)
            ind = slice(istart, istop)

            if interp_pix is None or interp_weights is None:
                p, w = hp.get_interp_weights(
                    self.nside, theta[ind], phi[ind], nest=self.nest
                )
            else:
                p = np.ascontiguousarray(interp_pix[:, ind])
                w = np.ascontiguousarray(interp_weights[:, ind])

            weights = np.ascontiguousarray(IQUweight[ind].T)

            buf = np.zeros(istop - istart, dtype=np.float64)
            fast_scanning_float32(buf, p, w, self._map_Q[:])
            if pol_deriv:
                signal[ind] = -2 * weights[2] * buf
            else:
                signal[ind] = weights[1] * buf

            buf[:] = 0
            fast_scanning_float32(buf, p, w, self._map_U[:])
            if pol_deriv:
                signal[ind] += 2 * weights[1] * buf
            else:
                signal[ind] += weights[2] * buf

            if not onlypol:
                if self._map is None:
                    raise RuntimeError("No temperature map to sample")
                buf[:] = 0
                fast_scanning_float32(buf, p, w, self._map[:])
                signal[ind] += weights[0] * buf

        return signal



#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""This does some simple tests of the TOAST runtime environment.
"""

import sys
import argparse
import traceback

from toast.mpi import get_world, Comm

from toast.utils import Logger, Environment, numba_threading_layer


def main():
    env = Environment.get()
    log = Logger.get()

    parser = argparse.ArgumentParser(
        description="Test the TOAST runtime environment.", fromfile_prefix_chars="@"
    )

    parser.add_argument(
        "--groupsize",
        required=False,
        type=int,
        default=0,
        help="size of processor groups used to distribute observations",
    )

    try:
        args = parser.parse_args()
    except SystemExit:
        return

    mpiworld, procs, rank = get_world()
    if rank == 0:
        print(env)
        log.info("Numba threading layer set to '{}'".format(numba_threading_layer))
    if mpiworld is None:
        log.info("Running serially with one process")
    else:
        if rank == 0:
            log.info("Running with {} processes".format(procs))

    groupsize = args.groupsize
    if groupsize <= 0:
        groupsize = procs

    if rank == 0:
        log.info("Using group size of {} processes".format(groupsize))

    comm = Comm(world=mpiworld, groupsize=groupsize)

    log.info(
        "Process {}:  world rank {}, group {} of {}, group rank {}".format(
            rank, comm.world_rank, comm.group + 1, comm.ngroups, comm.group_rank
        )
    )

    return


if __name__ == "__main__":
    try:
        main()
    except:
        # We have an unhandled exception on at least one process.  Print a stack
        # trace for this process and then abort so that all processes terminate.
        mpiworld, procs, rank = get_world()
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        lines = ["Proc {}: {}".format(rank, x) for x in lines]
        print("".join(lines), flush=True)
        if mpiworld is not None:
            mpiworld.Abort(6)
